import logging
import uuid

import gevent

from lymph.utils.gpool import NonBlockingPool


logger = logging.getLogger(__name__)


def get_trace(greenlet=None):
    greenlet = greenlet or gevent.getcurrent()
    if not hasattr(greenlet, '_lymph_trace'):
        greenlet._lymph_trace = {}
    return greenlet._lymph_trace


class GreenletWithTrace(gevent.Greenlet):

    def __init__(self, *args, **kwargs):
        super(GreenletWithTrace, self).__init__(*args, **kwargs)
        self._lymph_trace = get_trace().copy()


class Group(NonBlockingPool):
    greenlet_cls = GreenletWithTrace


def trace(**kwargs):
    get_trace().update(kwargs)


def set_id(trace_id=None):
    tid = trace_id or uuid.uuid4().hex
    trace(lymph_trace_id=tid)
    if trace_id is None:
        logger.debug('starting trace')
    return tid


def get_id():
    return get_trace().get('lymph_trace_id')


class TraceFormatter(logging.Formatter):
    def format(self, record):
        record.trace_id = get_id()
        return super(TraceFormatter, self).format(record)



import logging
import os
from functools import partial
import signal
import sys

import gevent
from setproctitle import setproctitle
import six

from lymph.utils import import_object, dump_stacks
from lymph.autoreload import set_source_change_callback
from lymph.cli.base import Command
from lymph.core.container import create_container
from lymph.utils.sockets import get_unused_port


logger = logging.getLogger(__name__)


SIGNAL_NAMES = {getattr(signal, name): name for name in dir(signal) if name.startswith('SIG') and '_' not in name}


def install_plugins(container, plugins):
    for name, plugin_config in six.iteritems(plugins):
        cls = import_object(plugin_config['class'])
        container.install_plugin(cls, **plugin_config)


def install_interfaces(container, interfaces):
    for name, instance_config in six.iteritems(interfaces):
        try:
            cls_name = instance_config['class']
        except KeyError:
            print("no instance class for '%s'" % name)
            sys.exit(1)
        cls = import_object(cls_name)
        instance = container.install_interface(cls, name=name)
        instance.apply_config(instance_config)


class InstanceCommand(Command):
    """
    Usage: lymph instance [--ip=<address> | --guess-external-ip | -g]
                         [--port <port> | -p <port>] [--reload] [--debug]
                         [--interface=<cls>]... [options]

    {INSTANCE_OPTIONS}

    {COMMON_OPTIONS}
    """

    proctitle = 'lymph-instance'
    short_description = 'Run a single service instance (one process).'

    def run(self):
        debug = self.args.get('--debug')

        self._setup_container(debug)

        if debug:
            self._start_backdoor_terminal()

        self._register_signals()

        self._set_process_title()

        self.container.start(register=not self.args.get('--isolated', False))

        if self.args.get('--reload'):
            set_source_change_callback(self.container.stop)

        self.container.join()

    def _setup_container(self, debug):
        self.container = create_container(self.config)
        self.container.debug = debug
        # Set global exception hook to send unhandled exception to the container's error_hook.
        sys.excepthook = self.container.excepthook

        install_plugins(self.container, self.config.get('plugins', {}))
        install_interfaces(self.container, self.config.get('interfaces', {}))

        for cls_name in self.args.get('--interface', ()):
            cls = import_object(cls_name)
            self.container.install_interface(cls)

    def _start_backdoor_terminal(self):
        # XXX(Mouad): Imported here since this is still broken in Python3.x
        from gevent.backdoor import BackdoorServer

        try:
            ip = self.config.get_raw('debug.backdoor_ip')
        except KeyError:
            ip = '127.0.0.1'
        port = get_unused_port()
        endpoint = '%s:%s' % (ip, port)

        banner = "Welcome to backdoor Terminal of %s" % self.container.service_name

        backdoor = BackdoorServer(
            endpoint,
            locals={'container': self.container, 'dump_stacks': dump_stacks},
            banner=banner)
        gevent.spawn(backdoor.serve_forever)

        self.container.backdoor_endpoint = endpoint

    def _register_signals(self):
        gevent.signal(signal.SIGINT, self._handle_termination_signal, signal.SIGINT)
        gevent.signal(signal.SIGTERM, self._handle_termination_signal, signal.SIGTERM)
        gevent.signal(signal.SIGQUIT, self._handle_termination_signal, signal.SIGQUIT, prehook=partial(dump_stacks, output=sys.stderr.write))

    def _handle_termination_signal(self, signalnum, prehook=None):
        logger.info('caught %s, pid=%s', SIGNAL_NAMES.get(signalnum, signalnum), os.getpid())
        if prehook:
            prehook()
        self.container.stop(signalnum=signalnum)
        self.container.join()
        sys.exit(0)

    def _set_process_title(self):
        setproctitle('lymph-instance (identity: %s, endpoint: %s, config: %s)' % (
            self.container.identity,
            self.container.endpoint,
            self.config.source,
        ))


class NodeCommand(InstanceCommand):
    """
    Usage: lymph node [--debug] [options]

    {INSTANCE_OPTIONS}

    {COMMON_OPTIONS}
    """

    proctitle = 'lymph-node'
    short_description = 'Run a node service that manages a group of processes on the same machine.'

    def run(self):
        self.config.update({
            'interfaces': {
                'node': {
                    'class': 'lymph.services.node:Node',
                    'instances': self.config.get_raw('instances', {}),
                    'sockets': self.config.get_raw('sockets', {}),
                },
            }
        })
        os.environ['LYMPH_NODE_CONFIG'] = self.config.source
        super(NodeCommand, self).run()




import sys
import gevent
import six


class Component(object):
    def __init__(self, error_hook=None, pool=None):
        self._parent_component = None
        self.__error_hook = error_hook
        self.__pool = pool

    def on_start(self):
        pass

    def on_stop(self, **kwargs):
        pass

    def _get_metrics(self):
        return []

    @property
    def pool(self):
        if self.__pool is not None:
            return self.__pool
        if not self._parent_component:
            raise AttributeError("root component without pool")
        return self._parent_component.pool

    @property
    def error_hook(self):
        if self.__error_hook:
            return self.__error_hook
        if not self._parent_component:
            raise AttributeError("root component without error_hook")
        return self._parent_component.error_hook

    def spawn(self, func, *args, **kwargs):
        def _inner():
            try:
                return func(*args, **kwargs)
            except gevent.GreenletExit:
                raise
            except:
                self.error_hook(sys.exc_info())
                raise
        return self.pool.spawn(_inner)


class Declaration(object):
    def __init__(self, factory):
        self.factory = factory

    def install(self, componentized):
        return componentized.install(self)

    def __call__(self, *args, **kwargs):
        return self.factory(*args, **kwargs)

    def __get__(self, componentized, cls):
        if componentized is None:
            return self
        try:
            return componentized._declared_components[self]
        except KeyError:
            return self.install(componentized)


class ComponentizedBase(type):
    def __new__(cls, clsname, bases, attrs):
        declarations = set()
        for base in bases:
            if isinstance(base, ComponentizedBase):
                declarations.update(base.declarations)
        for name, value in six.iteritems(attrs):
            if isinstance(value, Declaration):
                value.name = name
                declarations.add(value)
        new_cls = super(ComponentizedBase, cls).__new__(cls, clsname, bases, attrs)
        new_cls.declarations = declarations
        return new_cls


@six.add_metaclass(ComponentizedBase)
class Componentized(Component):
    def __init__(self, **kwargs):
        super(Componentized, self).__init__(**kwargs)
        self._declared_components = {}
        self.__all_components = []

    def add_component(self, component):
        component._parent_component = self
        self.__all_components.append(component)

    def install(self, factory, **kwargs):
        if factory in self._declared_components:
            raise RuntimeError("already installed: %s" % factory)
        component = factory(self, **kwargs)
        self._declared_components[factory] = component
        self.add_component(component)
        return component

    def on_start(self):
        for declaration in self.declarations:
            # FIXME: is this the right place to force declaration resolution?
            declaration.__get__(self, type(self))
        for component in self.__all_components:
            component.on_start()

    def on_stop(self, **kwargs):
        for component in reversed(self.__all_components):
            component.on_stop(**kwargs)

    def _get_metrics(self):
        for metric in super(Componentized, self)._get_metrics():
            yield metric
        for component in self.__all_components:
            for metric in component._get_metrics():
                yield metric



import lymph
from lymph.client import Client
from lymph.cli.base import Command


class SubscribeCommand(Command):
    """
    Usage: lymph subscribe <event-type>... [options]

    {COMMON_OPTIONS}
    """

    short_description = 'Prints events to stdout.'

    def run(self):
        event_type = self.args.get('<event-type>')

        class Subscriber(lymph.Interface):
            @lymph.event(*event_type)
            def on_event(self, event):
                print('%s: %r' % (event.evt_type, event.body))

        client = Client.from_config(self.config, interface_cls=Subscriber)
        client.container.join()




import socket
import os

from six.moves import urllib
import netifaces


def guess_external_ip():
    gateways = netifaces.gateways()
    try:
         ifnet = gateways['default'][netifaces.AF_INET][1]
         return netifaces.ifaddresses(ifnet)[netifaces.AF_INET][0]['addr']
    except (KeyError, IndexError):
        return


def bind_zmq_socket(sock, address, port=None):
    endpoint = address if '://' in address else 'tcp://%s' % address
    p = urllib.parse.urlparse(endpoint)
    if port and p.port and p.port != port:
        raise ValueError('two port numbers given: %s and %s' % (p.port, port))

    if p.port:
        sock.bind(endpoint)
        port = p.port
    elif port:
        endpoint = '%s:%s' % (endpoint, port)
        sock.bind(endpoint)
    else:
        port = sock.bind_to_random_port(endpoint)
        endpoint = '%s:%s' % (endpoint, port)
    return endpoint, port


# adapted from https://github.com/mozilla-services/chaussette/
def create_socket(host, family=socket.AF_INET, type=socket.SOCK_STREAM,
                  backlog=2048, blocking=True, inheritable=False):
    if family == socket.AF_UNIX and not host.startswith('unix:'):
        raise ValueError('Your host needs to have the unix:/path form')
    if host.startswith('unix:'):
        family = socket.AF_UNIX

    if host.startswith('fd://'):
        fd = int(host[5:])
        sock = socket.fromfd(fd, family, type)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    else:
        sock = socket.socket(family, type)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        if host.startswith('unix:'):
            filename = host[len('unix:'):]
            try:
                os.remove(filename)
            except OSError:
                pass
            sock.bind(filename)
        else:
            if ':' in host:
                host, port = host.rsplit(':', 1)
                port = int(port)
            else:
                host, port = '0.0.0.0', int(host)
            sock.bind((host, port))
        sock.listen(backlog)

    if blocking:
        sock.setblocking(1)
    else:
        sock.setblocking(0)

    # Required since Python 3.4 to be able to share a socket with a child
    # process.
    if inheritable and hasattr(os, 'set_inheritable'):
        os.set_inheritable(sock.fileno(), True)

    return sock


def get_unused_port(host="127.0.0.1", family=socket.AF_INET, socktype=socket.SOCK_STREAM):
    tempsock = socket.socket(family, socktype)
    tempsock.bind((host, 0))
    port = tempsock.getsockname()[1]
    tempsock.close()
    del tempsock
    return port




import functools
import json
import logging
import six

from kazoo.protocol.states import EventType, KazooState
from kazoo.handlers.gevent import SequentialGeventHandler
from kazoo.exceptions import NoNodeError, ConnectionLoss

from .base import BaseServiceRegistry
from lymph.exceptions import LookupFailure, RegistrationFailure
from lymph.utils.logging import setup_logger


logger = logging.getLogger(__name__)

DEFAULT_CHROOT = '/lymph'


class ZookeeperServiceRegistry(BaseServiceRegistry):
    def __init__(self, zkclient, pool=None):
        super(ZookeeperServiceRegistry, self).__init__(pool=pool)
        self.client = zkclient
        if not self.client.chroot:
            self.client.chroot = DEFAULT_CHROOT
        self.client.add_listener(self.on_kazoo_state_change)
        self.start_count = 0
        self.registered_names = {}

    @classmethod
    def from_config(cls, config, **kwargs):
        zkclient = config.get_instance('zkclient', handler=SequentialGeventHandler())
        return cls(zkclient=zkclient, **kwargs)

    def on_start(self, timeout=10):
        setup_logger('kazoo')
        self.start_count += 1
        if self.start_count > 1:
            return
        started = self.client.start_async()
        started.wait(timeout=timeout)
        if not self.client.connected:
            raise RuntimeError('could not connect to zookeeper')
        logger.debug('connected to zookeeper (version=%s)', '.'.join(map(str, self.client.server_version())))

    def on_stop(self, **kwargs):
        self.start_count -= 1
        if self.start_count != 0:
            return
        self.client.stop()

    def on_kazoo_state_change(self, state):
        logger.info('kazoo connection state changed to %s', state)
        if state == KazooState.CONNECTED:
            for name, instance in self.registered_names.items():
                self.spawn(self.register, name, instance)
            for service in six.itervalues(self.cache):
                self.spawn(self.lookup, service, timeout=None)

    def on_service_name_watch(self, service, event):
        try:
            self.lookup(service)
        except LookupFailure:
            pass
        except Exception:
            logger.exception('error in service type watcher')

    def on_service_watch(self, service, event):
        try:
            prefix, service_name, identity = event.path.rsplit('/', 2)
            if event.type == EventType.DELETED:
                service.remove(identity)
        except Exception:
            logger.exception('error in service watcher')

    def _get_service_znode(self, service, service_name, identity):
        path = self._get_zk_path(service_name, identity)
        result = self.client.get_async(
            path, watch=functools.partial(self.on_service_watch, service))
        value, znode = result.get()
        items = six.iteritems(json.loads(value.decode('utf-8')))
        return {str(k): str(v) for k, v in items}

    def discover(self):
        result = self.client.get_children_async(
            path='/services',
        )
        try:
            return list(result.get())
        except NoNodeError:
            return []

    def lookup(self, service, timeout=1):
        service_name = service.name
        result = self.client.get_children_async(
            path='/services/%s' % (service_name, ),
            watch=functools.partial(self.on_service_name_watch, service),
        )
        try:
            names = result.get(timeout=timeout)
        except NoNodeError:
            raise LookupFailure("failed to resolve %s" % service.name)
        except ConnectionLoss:
            logger.warning("lost zookeeper connection")
            return service
        logger.info("lookup %s %r", service_name, names)
        identities = set(service.identities())
        for name in names:
            kwargs = self._get_service_znode(service, service_name, name)
            identity = kwargs.pop('identity')
            service.update(identity, **kwargs)
            try:
                identities.remove(identity)
            except KeyError:
                pass
        for identity in identities:
            service.remove(identity)
        return service

    def _get_zk_path(self, service_name, identity):
        return '/services/%s/%s' % (service_name, identity)

    def register(self, service_name, instance, timeout=1):
        path = self._get_zk_path(service_name, instance.identity)
        value = json.dumps(instance.serialize())

        result = self.client.create_async(
            path,
            value.encode('utf-8'),
            ephemeral=True, makepath=True)
        # FIXME: result.set_exception(RegistrationFailure())
        result.get(timeout=timeout)
        self.registered_names[service_name] = instance

    def unregister(self, service_name, instance, timeout=1):
        path = self._get_zk_path(service_name, instance.identity)
        result = self.client.delete_async(path)
        result.set_exception(RegistrationFailure())
        result.get(timeout=timeout)
        del self.registered_names[service_name]



from unittest import TestCase

from lymph.utils import import_object, Undefined


class ImportTests(TestCase):

    def test_import_object(self):
        from lymph.core.container import ServiceContainer
        cls = import_object('lymph.core.container:ServiceContainer')
        self.assertIs(cls, ServiceContainer)

    def test_import_object_without_colon(self):
        self.assertRaises(ValueError, import_object, 'lymph.core.container.ServiceContainer')
        self.assertRaises(ValueError, import_object, 'lymph.core.container')


class UndefinedTests(TestCase):
    def test_properties(self):
        self.assertNotEqual(Undefined, None)
        self.assertNotEqual(Undefined, False)
        self.assertFalse(bool(Undefined))
        self.assertEqual(str(Undefined), 'Undefined')



from __future__ import absolute_import, unicode_literals

import functools

import newrelic.agent

from lymph.core import trace
from lymph.core.plugins import Plugin
from lymph.core.container import ServiceContainer
from lymph.web.interfaces import WebServiceInterface


def with_trace_id(func):
    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        newrelic.agent.add_custom_parameter('trace_id', trace.get_id())
        return func(*args, **kwargs)
    return wrapped


class NewrelicPlugin(Plugin):
    def __init__(self, container, config_file=None, environment=None, **kwargs):
        super(NewrelicPlugin, self).__init__()
        self.container = container
        self.container.error_hook.install(self.on_error)
        self.container.http_request_hook.install(self.on_http_request)
        newrelic.agent.initialize(config_file, environment)
        for method in ('send_request', 'emit_event', 'lookup'):
            setattr(ServiceContainer, method, newrelic.agent.function_trace()(getattr(ServiceContainer, method)))

    def on_interface_installation(self, interface):
        for name, method in interface.methods.items():
            method.decorate(with_trace_id)
            method.decorate(newrelic.agent.background_task())
        if isinstance(interface, WebServiceInterface):
            interface.application = newrelic.agent.wsgi_application()(interface.application)

    def on_error(self, exc_info, **kwargs):
        newrelic.agent.add_custom_parameter('trace_id', trace.get_id())
        newrelic.agent.record_exception(exc_info)

    def on_http_request(self, request, rule, kwargs):
        newrelic.agent.set_transaction_name("%s %s" % (request.method, rule))
        newrelic.agent.add_custom_parameter('trace_id', trace.get_id())



import unittest

import gevent

from lymph.utils import gpool


def some_work():
    pass


class NonBlockingPoolTestCase(unittest.TestCase):

    def setUp(self):
        self.pool = gpool.NonBlockingPool(size=2)

    def test_full_pool_status(self):
        self.pool.spawn(some_work)
        self.pool.spawn(some_work)

        self.assertTrue(self.pool.full())
        self.assertEqual(self.pool.free_count(), 0)

    def test_spawn_on_full_pool_should_fail(self):
        self.pool.spawn(some_work)
        self.pool.spawn(some_work)

        with self.assertRaises(gpool.RejectExcecutionError):
            self.pool.spawn(some_work)

    def test_finished_work_should_freeup_resources_from_pool(self):
        self.pool.spawn(some_work)

        gevent.wait(self.pool)

        self.assertEqual(self.pool.free_count(), 2)



import unittest


class InitializeTest(unittest.TestCase):

    def test_assure_that_lymph_is_initialized_by_testrunner(self):
        import lymph.monkey
        self.assertTrue(lymph.monkey.patch._initialized)



import logging

from gevent.pywsgi import WSGIServer, WSGIHandler


logger = logging.getLogger(__name__)


class LymphWSGIHandler(WSGIHandler):

    def format_request(self):
        # XXX(Mouad): Copied shamessly from gevent.pywsgi.WSGIHandler's format_request
        # and removed only the datetime from the output, since it's already part of
        # lymph logger format.
        length = self.response_length or '-'
        if self.time_finish:
            delta = '%f' % (self.time_finish - self.time_start)
        else:
            delta = '-'
        client_address = self.client_address[0] if isinstance(self.client_address, tuple) else self.client_address
        return 'client=%s - - "%s" status=%s length=%s duration=%s (seconds)' % (
            client_address or '-',
            getattr(self, 'requestline', ''),
            (getattr(self, 'status', None) or '000').split()[0],
            length,
            delta)

    def log_request(self):
        # XXX(Mouad): Workaround to log correctly in gevent wsgi.
        # https://github.com/gevent/gevent/issues/106
        logger.info(self.format_request())


class LymphWSGIServer(WSGIServer):
    handler_class = LymphWSGIHandler



import datetime
import decimal
import json
import unittest
import uuid

from lymph.serializers import base
from lymph.utils import Undefined


class SerializerBaseTest(unittest.TestCase):
    def setUp(self):
        self.json_serializer = base.BaseSerializer(dumps=json.dumps, loads=json.loads, dump=json.dump, load=json.load)

    def assertJsonEquals(self, a, b):
        self.assertEquals(json.loads(a), b)

    def test_DatetimeSerializer_serialize(self):
        serializer = base.DatetimeSerializer()
        self.assertEqual(
            serializer.serialize(datetime.datetime(1900, 1, 1, 0, 0, 0, 1)),
            "1900-01-01T00:00:00Z")
        self.assertEqual(
            serializer.serialize(datetime.datetime(2014, 9, 12, 8, 33, 12, 34)),
            "2014-09-12T08:33:12Z")

    def test_DatetimeSerializer_deserialize(self):
        serializer = base.DatetimeSerializer()
        self.assertEqual(serializer.deserialize("1900-01-01T00:00:00Z"),
                         datetime.datetime(1900, 1, 1, 0, 0))
        self.assertEqual(serializer.deserialize("2014-09-12T08:33:12Z"),
                         datetime.datetime(2014, 9, 12, 8, 33, 12))
        self.assertRaises(ValueError, serializer.deserialize, "2014-09-12T08:33:12")
        self.assertRaises(ValueError, serializer.deserialize, "2014-09-12T25:33:12Z")
        self.assertRaises(ValueError, serializer.deserialize, "2014-02-30T08:33:12Z")

    def test_DateSerializer_serialize(self):
        serializer = base.DateSerializer()
        self.assertEqual(
            serializer.serialize(datetime.date(1900, 1, 1)), "1900-01-01")
        self.assertEqual(
            serializer.serialize(datetime.date(2014, 9, 12)), "2014-09-12")

    def test_DateSerializer_deserialize(self):
        serializer = base.DateSerializer()
        self.assertEqual(serializer.deserialize("1900-01-01"),
                         datetime.date(1900, 1, 1))
        self.assertEqual(serializer.deserialize("2014-09-12"),
                         datetime.date(2014, 9, 12))
        self.assertRaises(ValueError, serializer.deserialize, "2014-02-30")

    def test_TimeSerializer_serialize(self):
        serializer = base.TimeSerializer()
        self.assertEqual(
            serializer.serialize(datetime.time(0, 0, 0, 1)), "00:00:00Z")
        self.assertEqual(
            serializer.serialize(datetime.time(8, 33, 12, 34)), "08:33:12Z")

    def test_TimeSerializer_deserialize(self):
        serializer = base.TimeSerializer()
        self.assertEqual(serializer.deserialize("00:00:00Z"),
                         datetime.time(0, 0, 0))
        self.assertEqual(serializer.deserialize("18:33:12Z"),
                         datetime.time(18, 33, 12))
        self.assertRaises(ValueError, serializer.deserialize, "25:33:12")

    def test_decimal_serialize(self):
        serializer = base.StrSerializer(decimal.Decimal)
        self.assertEqual(serializer.serialize(decimal.Decimal('3.1415')), "3.1415")
        self.assertEqual(serializer.serialize(decimal.Decimal('2.50')), "2.50")
        self.assertEqual(serializer.serialize(decimal.Decimal('NaN')), "NaN")
        self.assertEqual(serializer.serialize(decimal.Decimal('Infinity')), "Infinity")

    def test_decimal_deserialize(self):
        serializer = base.StrSerializer(decimal.Decimal)
        self.assertEqual(serializer.deserialize("3.1415"), decimal.Decimal('3.1415'))
        self.assertEqual(serializer.deserialize("2.50"), decimal.Decimal('2.50'))
        self.assertEqual(serializer.deserialize("NaN").number_class(), 'NaN')
        self.assertEqual(serializer.deserialize("Infinity"), decimal.Decimal('Infinity'))
        self.assertRaises(decimal.InvalidOperation, serializer.deserialize, "foo")

    def test_SetSerializer_serialize(self):
        serializer = base.SetSerializer()
        self.assertEqual(
            set(serializer.serialize(set(["as", "this", "is", "a", "set"]))),
            set(['this', 'a', 'as', 'set', 'is']))

    def test_SetSerializer_deserialize(self):
        serializer = base.SetSerializer()
        self.assertEqual(serializer.deserialize(["as", "this", "is", "a", "set"]),
                         set(["as", "this", "is", "a", "set"]))

    def test_BaseSerializer_dump_object(self):
        serializer = base.BaseSerializer(dumps=json.dumps, loads=json.loads, dump=json.dump, load=json.load)
        self.assertEqual(serializer.dump_object(datetime.datetime(2014, 9, 12, 8, 33, 12, 34)),
                         {'__type__': 'datetime', '_': '2014-09-12T08:33:12Z'})
        self.assertEqual(serializer.dump_object(datetime.date(2014, 9, 12)),
                         {'__type__': 'date', '_': '2014-09-12'})
        self.assertEqual(serializer.dump_object(decimal.Decimal('3.1415')),
                         {'__type__': 'Decimal', '_': '3.1415'})
        self.assertEqual(
            serializer.dump_object(uuid.UUID('00000000-0000-4000-8000-000000000000')),
            {'__type__': 'UUID', '_': '00000000-0000-4000-8000-000000000000'},
        )

    def test_BaseSerializer_dump(self):
        serializer = base.BaseSerializer(dumps=json.dumps, loads=json.loads, dump=json.dump, load=json.load)
        self.assertJsonEquals(
            serializer.dumps(datetime.datetime(2014, 9, 12, 8, 33, 12, 34)),
            {"__type__": "datetime", "_": "2014-09-12T08:33:12Z"}
        )
        self.assertJsonEquals(
            serializer.dumps(decimal.Decimal('NaN')),
            {"__type__": "Decimal", "_": "NaN"}
        )
        self.assertJsonEquals(
            serializer.dumps(set([datetime.datetime(2014, 9, 12, 8, 33, 12, 34)])),
            {"__type__": "set", "_": [{"__type__": "datetime", "_": "2014-09-12T08:33:12Z"}]}
        )

    def test_BaseSerializer_load_object(self):
        serializer = base.BaseSerializer(dumps=json.dumps, loads=json.loads, dump=json.dump, load=json.load)
        self.assertEqual(serializer.load_object(
                         {'__type__': 'datetime', '_': '2014-09-12T08:33:12Z'}),
                         datetime.datetime(2014, 9, 12, 8, 33, 12))
        self.assertEqual(serializer.load_object(
                         {'__type__': 'date', '_': '2014-09-12'}),
                         datetime.date(2014, 9, 12))
        self.assertEqual(serializer.load_object(
                         {'__type__': 'Decimal', '_': '3.1415'}),
                         decimal.Decimal('3.1415'))
        self.assertEqual(serializer.load_object(
                         {'__type__': 'Decimal', '_': 'NaN'}).number_class(),
                         'NaN')
        self.assertEqual(
            serializer.load_object({'__type__': 'set', '_': ['this', 'a', 'as', 'set', 'is']}),
            set(['this', 'a', 'as', 'set', 'is']),
        )
        self.assertEqual(
            serializer.load_object({'__type__': 'UUID', '_': '00000000-0000-4000-8000-000000000000'}),
            uuid.UUID('00000000-0000-4000-8000-000000000000'),
        )

    def test_BaseSerializer_loads(self):
        serializer = self.json_serializer
        normal_datetime = '{"__type__": "datetime", "_": "2014-09-12T08:33:12Z"}'
        normal_date = '{"__type__": "date", "_": "2014-09-12"}'
        number = '{"__type__": "Decimal", "_": "3.1415"}'
        nan = '{"__type__": "Decimal", "_": "NaN"}'
        some_set = '{"__type__": "set", "_": ["this", "a", "as", "set", "is"]}'

        self.assertEqual(serializer.loads(normal_datetime),
                         datetime.datetime(2014, 9, 12, 8, 33, 12))
        self.assertEqual(serializer.loads(normal_date),
                         datetime.date(2014, 9, 12))
        self.assertEqual(serializer.loads(number), decimal.Decimal('3.1415'))
        self.assertEqual(serializer.loads(nan).number_class(), 'NaN')
        self.assertEqual(serializer.loads(some_set),
                         set([u'this', u'a', u'as', u'set', u'is']))
        self.assertEqual(serializer.loads('{"__type__": "set", "_": ['
                         '{"__type__": "datetime", "_": "2014-09-12T08:33:12Z"}, '
                         '{"__type__": "datetime", "_": "2015-09-12T08:33:12Z"}'
                         ']}'), set([datetime.datetime(2014, 9, 12, 8, 33, 12),
                                     datetime.datetime(2015, 9, 12, 8, 33, 12)]))

    def test_undefined(self):
        self.assertJsonEquals(self.json_serializer.dumps(Undefined), {'__type__': 'UndefinedType', '_': ''})
        self.assertIs(self.json_serializer.loads('{"__type__": "UndefinedType", "_": ""}'), Undefined)



import textwrap

from lymph.client import Client
from lymph.cli.base import Command, handle_request_errors


class InspectCommand(Command):
    """
    Usage: lymph inspect [--ip=<address> | --guess-external-ip | -g] <address> [options]

    Options:
      --ip=<address>               Use this IP for all sockets.
      --guess-external-ip, -g      Guess the public facing IP of this machine and
                                   use it instead of the provided address.

    {COMMON_OPTIONS}

    """

    short_description = 'Describes the RPC interface of a service'

    @handle_request_errors
    def run(self):
        address = self.args['<address>']
        client = Client.from_config(self.config)
        result = client.request(address, 'lymph.inspect', {}, timeout=5).body

        print('RPC interface of {}\n'.format(self.terminal.bold(address)))

        for method in sorted(result['methods'], key=lambda m: m['name']):
            print(
                "rpc {name}({params})\n\t {help}\n".format(
                    name=self.terminal.red(method['name']),
                    params=self.terminal.yellow(', '.join(method['params'])),
                    help='\n    '.join(textwrap.wrap(method['help'], 70)),
                )
            )




from __future__ import absolute_import

import sys
from contextlib import contextmanager

import logging
import kombu
import kombu.mixins
import kombu.pools

from lymph.events.base import BaseEventSystem
from lymph.core.events import Event
from lymph.utils.logging import setup_logger


logger = logging.getLogger(__name__)


DEFAULT_SERIALIZER = 'lymph-msgpack'
DEFAULT_EXCHANGE = 'lymph'


class EventConsumer(kombu.mixins.ConsumerMixin):
    def __init__(self, event_system, connection, queue, handler):
        self.connection = connection
        self.queue = queue
        self.handler = handler
        self.greenlet = None
        self.event_system = event_system

    def get_consumers(self, Consumer, channel):
        return [Consumer(queues=[self.queue], callbacks=[self.on_kombu_message])]

    def create_connection(self):
        return kombu.pools.connections[self.connection].acquire(block=True)

    def on_kombu_message(self, body, message):
        logger.debug("received kombu message %r", body)

        def message_handler():
            try:
                event = Event.deserialize(body)
                self.handler(event)
                message.ack()
            except:
                logger.exception('failed to handle event from queue %r', self.handler.queue_name)
                # FIXME: add requeue support here
                message.reject()
                # Since the message handler can be run sequentially, we are catching all exception
                # and reporting them here.
                self.event_system.container.error_hook(sys.exc_info())
            finally:
                if self.handler.once:
                    self.event_system.unsubscribe(self.handler)

        if self.handler.sequential:
            message_handler()
        else:
            self.event_system.container.spawn(message_handler)

    def start(self):
        if self.greenlet:
            return
        self.should_stop = False
        self.greenlet = self.event_system.container.spawn(self.run)

    def stop(self, **kwargs):
        if not self.greenlet:
            return
        self.should_stop = True
        self.greenlet.join()
        self.greenlet = None


class KombuEventSystem(BaseEventSystem):
    def __init__(self, connection, exchange_name, serializer=DEFAULT_SERIALIZER):
        super(KombuEventSystem, self).__init__()
        self.connection = connection
        self.exchange = kombu.Exchange(exchange_name, 'topic', durable=True)
        self.waiting_exchange = kombu.Exchange('%s_waiting' % exchange_name, 'direct', durable=True)
        self.waiting_queues = {}
        self.serializer = serializer
        self.consumers_by_queue = {}

    def on_start(self):
        setup_logger('kombu')

    def on_stop(self, **kwargs):
        for consumer in self.consumers_by_queue.values():
            consumer.stop(**kwargs)
        self.consumers_by_queue.clear()

    @classmethod
    def from_config(cls, config, **kwargs):
        exchange_name = config.get('exchange', DEFAULT_EXCHANGE)
        serializer = config.get('serializer', DEFAULT_SERIALIZER)
        connection = kombu.Connection(**config)
        return cls(connection, exchange_name, serializer=serializer, **kwargs)

    def setup_consumer(self, handler):
        with self._get_connection() as conn:
            self.exchange(conn).declare()
            if handler.broadcast:
                queue = kombu.Queue(handler.queue_name, auto_delete=True, durable=False)
            else:
                queue = kombu.Queue(handler.queue_name, durable=True, auto_delete=handler.once)
            queue(conn).declare()
            for event_type in handler.event_types:
                queue(conn).bind_to(exchange=self.exchange, routing_key=event_type)
        consumer = EventConsumer(self, self.connection, queue, handler)
        self.consumers_by_queue[handler.queue_name] = consumer
        return consumer

    def subscribe(self, handler, consume=True):
        try:
            consumer = self.consumers_by_queue[handler.queue_name]
        except KeyError:
            consumer = self.setup_consumer(handler)
        else:
            if consumer.handler != handler:
                raise RuntimeError('cannot subscribe to queue %r more than once' % handler.queue_name)
        if consume:
            consumer.start()
        return consumer

    def unsubscribe(self, handler):
        queue_name = handler.queue_name
        try:
            consumer = self.consumers_by_queue[queue_name]
        except KeyError:
            raise KeyError('there is no subscription for %r' % queue_name)
        if consumer.handler != handler:
            raise KeyError('%s is not subscribed to %r' % (handler, queue_name))
        consumer.stop()
        del self.consumers_by_queue[queue_name]

    @contextmanager
    def _get_connection(self):
        with kombu.pools.connections[self.connection].acquire(block=True) as conn:
            yield conn

    def _get_waiting_queue(self, conn, event_type, delay):
        delay_ms = int(1000 * delay)
        queue_name = '%s-wait_%s' % (event_type, delay_ms)
        try:
            return self.waiting_queues[queue_name]
        except KeyError:
            pass
        queue = kombu.Queue(queue_name, durable=True, queue_arguments={
            'x-dead-letter-exchange': self.exchange.name,
            'x-dead-letter-routing-key': event_type,
            'x-message-ttl': delay_ms,
        })
        self.waiting_exchange(conn).declare()
        queue(conn).declare()
        queue(conn).bind_to(exchange=self.waiting_exchange, routing_key=queue_name)
        self.waiting_queues[queue_name] = queue
        return queue

    def emit(self, event, delay=0):
        with self._get_connection() as conn:
            producer = conn.Producer(serializer=self.serializer)
            if delay:
                queue = self._get_waiting_queue(conn, event.evt_type, delay)
                routing_key = queue.name
                exchange = self.waiting_exchange
            else:
                routing_key = event.evt_type
                exchange = self.exchange
            producer.publish(event.serialize(), routing_key=routing_key, exchange=exchange, declare=[self.exchange])



import collections


class RawMetric(object):

    def __init__(self, name, value, tags=None):
        self._name = name
        self._value = value
        self._tags = tags or {}

    def __iter__(self):
        yield (self._name, self._value, self._tags)

    def __repr__(self):
        return '%s(name=%r, value=%r, tags=%r)' % (
            self.__class__.__name__,
            self._name,
            self._value,
            self._tags)

    __str__ = __repr__


class Counter(RawMetric):

    def __init__(self, name, tags=None):
        super(Counter, self).__init__(name, 0, tags)

    def __iadd__(self, value):
        self._value += value
        return self


class TaggedCounter(RawMetric):

    def __init__(self, name, tags=None):
        super(TaggedCounter, self).__init__(name, collections.Counter(), tags)

    def incr(self, _value=1, **tags):
        tags.update(self._tags)
        self._value[frozenset(tags.items())] += _value

    def __iter__(self):
        for tags, count in self._value.items():
            yield self._name, count, dict(tags)



import gevent

from lymph.core.events import EventDispatcher
from lymph.events.base import BaseEventSystem


class LocalEventSystem(BaseEventSystem):
    def __init__(self, **kwargs):
        super(LocalEventSystem, self).__init__(**kwargs)
        self.dispatcher = EventDispatcher()

    def subscribe(self, handler, **kwargs):
        for event_type in handler.event_types:
            self.dispatcher.register(event_type, handler)

    def unsubscribe(self, handler):
        raise NotImplementedError()

    def emit(self, event, delay=0):
        if delay:
            gevent.spawn_later(delay, self.dispatcher, event)
        else:
            self.dispatcher(event)



from __future__ import absolute_import

import logging

from nose.plugins import Plugin

log = logging.getLogger('nose.plugins.lymph')


class LymphPlugin(Plugin):
    """
    Initializes the lymph framework before tests are run
    """

    name = 'lymph'

    def begin(self):
        log.info("Initializing lymph framework")
        import lymph.monkey
        lymph.monkey.patch()



import logging

from lymph.core.components import Component


logger = logging.getLogger(__name__)


class Plugin(Component):
    def on_interface_installation(self, interface):
        pass


class Hook(object):
    def __init__(self, name='hook'):
        self.name = name
        self.callbacks = []

    def install(self, callback):
        self.callbacks.append(callback)

    def __call__(self, *args, **kwargs):
        for callback in self.callbacks:
            try:
                callback(*args, **kwargs)
            except:
                logger.exception('%s failure', self.name)



import abc
import six

from lymph.core.services import Service
from lymph.core.components import Component


@six.add_metaclass(abc.ABCMeta)
class BaseServiceRegistry(Component):
    def __init__(self, **kwargs):
        super(BaseServiceRegistry, self).__init__(**kwargs)
        self.cache = {}

    def get(self, service_name, **kwargs):
        try:
            service = self.cache[service_name]
        except KeyError:
            service = Service(name=service_name)
            self.lookup(service, **kwargs)
            self.cache[service_name] = service
        return service

    @abc.abstractmethod
    def discover(self):
        raise NotImplementedError

    @abc.abstractmethod
    def lookup(self, service, timeout=1):
        raise NotImplementedError

    @abc.abstractmethod
    def register(self, service_name, instance):
        raise NotImplementedError

    @abc.abstractmethod
    def unregister(self, service_name, instance):
        raise NotImplementedError



import abc
import unittest
import warnings

import gevent
import random
import six
import mock

from kazoo.handlers.gevent import SequentialGeventHandler
from kazoo.testing.harness import KazooTestHarness

from lymph.core.container import ServiceContainer
from lymph.core.connection import Connection
from lymph.core.interfaces import Interface, Proxy
from lymph.core.rpc import ZmqRPCServer
from lymph.core.messages import Message
from lymph.discovery.static import StaticServiceRegistryHub
from lymph.events.local import LocalEventSystem
from lymph.exceptions import RemoteError
from lymph.client import Client
from lymph.utils.sockets import get_unused_port, create_socket

import werkzeug.test
from werkzeug.wrappers import BaseResponse

from lymph.testing.mock_helpers import RpcMockTestCase, EventMockTestCase


class MockServiceNetwork(object):
    def __init__(self):
        self.service_containers = {}
        self.next_port = 1
        self.discovery_hub = StaticServiceRegistryHub()
        self.events = LocalEventSystem()

    def add_service(self, cls, interface_name=None, **kwargs):
        port = self.next_port
        self.next_port += 1
        registry = self.discovery_hub.create_registry()
        container = MockServiceContainer(
            registry=registry,
            events=self.events,
            rpc=MockRPCServer(ip='300.0.0.1', port=port, mock_network=self),
            **kwargs)
        container.install_interface(cls, name=interface_name)
        self.service_containers[container.endpoint] = container
        container._mock_network = self
        return container

    def start(self):
        for container in six.itervalues(self.service_containers):
            container.start()

    def stop(self, **kwargs):
        for container in six.itervalues(self.service_containers):
            container.stop()

    def join(self):
        for container in six.itervalues(self.service_containers):
            container.join()


class MockRPCServer(ZmqRPCServer):
    def __init__(self, *args, **kwargs):
        self.__mock_network = kwargs.pop('mock_network')
        super(MockRPCServer, self).__init__(*args, **kwargs)
        self._bind()

    def _bind(self):
        self.endpoint = 'mock://%s:%s' % (self.ip, self.port)

    def _close_sockets(self):
        pass

    def connect(self, endpoint):
        if endpoint not in self.connections:
            self.connections[endpoint] = Connection(self, endpoint)
        return self.connections[endpoint]

    def _send_message(self, endpoint, msg):
        dst = self.__mock_network.service_containers[endpoint]

        # Exercise the msgpack packing and unpacking.
        frames = msg.pack_frames()
        frames.insert(0, self.endpoint.encode('utf-8'))
        msg = Message.unpack_frames(frames)

        dst.server.recv_message(msg)

    def _recv_loop(self):
        pass


class MockServiceContainer(ServiceContainer):
    def __init__(self, *args, **kwargs):
        super(MockServiceContainer, self).__init__(*args, **kwargs)
        self.__shared_sockets = {}

    def get_shared_socket_fd(self, port):
        try:
            return self.__shared_sockets[port].fileno()
        except KeyError:
            host_port = get_unused_port()
            sock = create_socket('127.0.0.1:%s' % host_port, inheritable=True)
            self.__shared_sockets[port] = sock
            return sock.fileno()


class LymphIntegrationTestCase(KazooTestHarness):
    use_zookeeper = False

    def setUp(self):
        super(LymphIntegrationTestCase, self).setUp()
        self._containers = []
        if self.use_zookeeper:
            self.setup_zookeeper(handler=SequentialGeventHandler())

    def tearDown(self):
        super(LymphIntegrationTestCase, self).tearDown()
        for container in self._containers:
            container.stop()
        for container in self._containers:
            container.join()
        if self.use_zookeeper:
            self.teardown_zookeeper()
        self._containers = []

    def create_client(self, **kwargs):
        container, interface = self.create_container(**kwargs)
        return Client(container)

    def create_registry(self, **kwargs):
        return self.registry

    def create_event_system(self, **kwargs):
        return self.events

    def create_container(self, interface_cls=None, interface_name=None, events=None, registry=None, **kwargs):
        if not events:
            events = self.create_event_system(**kwargs)
        if not registry:
            registry = self.create_registry(**kwargs)
        container = ServiceContainer(
            events=events,
            registry=registry,
            rpc=ZmqRPCServer(),
            **kwargs)
        interface = None
        if interface_cls:
            interface = container.install_interface(interface_cls, name=interface_name)
        container.start()
        self._containers.append(container)
        return container, interface


class ClientInterface(Interface):
    service_type = 'client'


class LymphServiceTestCase(unittest.TestCase):
    client_class = ClientInterface
    client_name = 'client'
    client_config = {}
    service_class = ClientInterface
    service_name = 'client'
    service_config = {}

    def setUp(self):
        warnings.warn("deprecated, please use either RPCServiceTestCase or WebServiceTestCase")
        super(LymphServiceTestCase, self).setUp()
        self.network = MockServiceNetwork()
        self.service_container = self.network.add_service(
            self.service_class,
            interface_name=self.service_name
        )
        self.service = self.service_container.installed_interfaces[
            self.service_name
        ]
        self.service.apply_config(self.service_config)
        self.client_container = self.network.add_service(
            self.client_class,
            interface_name=self.client_name
        )
        self.client = self.client_container.installed_interfaces[
            self.client_name
        ]
        self.client.apply_config(self.client_config)
        self.network.start()

    def tearDown(self):
        super(LymphServiceTestCase, self).tearDown()
        self.network.stop()
        self.network.join()


@six.add_metaclass(abc.ABCMeta)
class RPCServiceTestCase(unittest.TestCase):
    service_config = {}

    def setUp(self):
        super(RPCServiceTestCase, self).setUp()
        self.network = MockServiceNetwork()

        self.container = self.network.add_service(self.service_class, interface_name=self.service_name)
        self.service = self.container.installed_interfaces[self.service_name]
        self.service.apply_config(self.service_config)

        self.network.start()

    @abc.abstractproperty
    def service_class(self):
        pass

    @property
    def service_name(self):
        return self.service_class.__name__

    def get_proxy(self, **kwargs):
        return Proxy(self.container, self.service_name, **kwargs)

    client = property(get_proxy)

    def request(self, *args, **kwargs):
        channel = self.container.send_request(self.service_name, *args, **kwargs)
        return channel.get(timeout=kwargs.get('timeout', 1))

    def emit(self, *args, **kwargs):
        return self.container.emit_event(*args, **kwargs)

    def tearDown(self):
        super(RPCServiceTestCase, self).tearDown()
        self.network.stop()
        self.network.join()


class WebServiceTestCase(RPCServiceTestCase):

    def setUp(self):
        # XXX(Mouad): Gevent master fail n python3 with: name 'dup' is not defined.
        # A workaround is to mock lymph.utils.sockets.create_socket to return
        # a dummy socket for us.
        port = get_unused_port()
        sock = create_socket('127.0.0.1:%s' % port, inheritable=True)
        with mock.patch('lymph.utils.sockets.create_socket', return_value=sock):
            super(WebServiceTestCase, self).setUp()

    @property
    def client(self):
        return werkzeug.test.Client(self.service.application, BaseResponse)


class AsyncTestsMixin(object):
    def wait_for(self, condition, timeout=2):
        with gevent.Timeout(timeout):
            while not condition():
                gevent.sleep(0)  # yield

    def assert_eventually_true(self, condition, message=None, timeout=2):
        if message is None:
            message = "The expected condition didn't happen within %s seconds" % timeout
        try:
            self.wait_for(condition, timeout=timeout)
        except gevent.Timeout:
            self.fail(message)

    def assert_temporarily_true(self, condition, message=None, timeout=2):
        try:
            with gevent.Timeout(timeout):
                while condition():
                    gevent.sleep(0)  # yield
        except gevent.Timeout:
            return
        if message is None:
            message = "The condition wasn't true for %s seconds" % timeout
        self.fail(message)



import textwrap
import functools

import six

from lymph.core.decorators import rpc, RPCBase
from lymph.exceptions import RemoteError, EventHandlerTimeout, Timeout, Nack
from lymph.core.components import Component, Componentized, ComponentizedBase
from lymph.core.monitoring import metrics

import gevent
from gevent.event import AsyncResult


REQUEST_TIMEOUT = 3  # seconds.


class AsyncResultWrapper(object):
    def __init__(self, container, handler, async_result):
        self.container = container
        self.handler = handler
        self.result = async_result

    def get(self, timeout=REQUEST_TIMEOUT):
        try:
            return self.result.get(timeout=timeout)
        except gevent.Timeout:
            self.container.unsubscribe(self.handler)
            raise EventHandlerTimeout


class InterfaceBase(ComponentizedBase):
    def __new__(cls, clsname, bases, attrs):
        methods = {}
        for base in bases:
            if isinstance(base, InterfaceBase):
                methods.update(base.methods)
        for name, value in six.iteritems(attrs):
            if isinstance(value, RPCBase):
                methods[name] = value
        new_cls = super(InterfaceBase, cls).__new__(cls, clsname, bases, attrs)
        new_cls.methods = methods
        return new_cls


class ProxyMethod(object):
    def __init__(self, func):
        self.func = func

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    def defer(self, *args, **kwargs):
        result = AsyncResult()
        gevent.spawn(self, *args, **kwargs).link(result)
        return result


class Proxy(Component):
    def __init__(self, container, address, timeout=REQUEST_TIMEOUT, namespace='', error_map=None):
        super(Proxy, self).__init__()
        self._container = container
        self._address = address
        self._method_cache = {}
        self._timeout = timeout
        self._namespace = namespace or address
        self._error_map = error_map or {}

        self.timeout_counts = metrics.Counter('timeout', {'address': address})
        self.exception_counts = metrics.TaggedCounter('exceptions', {'address': address})

    def _call(self, __name, **kwargs):
        channel = self._container.send_request(self._address, __name, kwargs)
        try:
            return channel.get(timeout=self._timeout).body
        except RemoteError as e:
            error_type = str(e.__class__)
            self.exception_counts.incr(name=e.__class__.__name__)
            if error_type in self._error_map:
                raise self._error_map[error_type]()
            raise
        except Timeout:
            self.timeout_counts += 1
            raise
        except Nack:
            self.exception_counts.incr(name='nack')
            raise

    def __getattr__(self, name):
        try:
            return self._method_cache[name]
        except KeyError:
            method = ProxyMethod(functools.partial(self._call, '%s.%s' % (self._namespace, name)))
            self._method_cache[name] = method
            return method

    def _get_metrics(self):
        yield self.timeout_counts
        yield self.exception_counts


@six.add_metaclass(InterfaceBase)
class Interface(Componentized):
    register_with_coordinator = True

    def __init__(self, container, name=None):
        super(Interface, self).__init__()
        self.container = container
        self.name = name or self.__class__.__name__

    def apply_config(self, config):
        self.config = config

    def get_description(self):
        return {}

    def handle_request(self, func_name, channel):
        self.methods[func_name].rpc_call(self, channel, **channel.request.body)

    def request(self, address, subject, body, timeout=REQUEST_TIMEOUT):
        channel = self.container.send_request(address, subject, body)
        return channel.get(timeout=timeout)

    def emit(self, event_type, payload, delay=0):
        self.container.emit_event(event_type, payload, delay=delay)

    def proxy(self, address, **kwargs):
        return Proxy(self.container, address, **kwargs)

    def subscribe(self, *event_types, **kwargs):
        def decorator(func):
            from lymph.core.events import EventHandler
            handler = EventHandler(self, func, event_types, **kwargs)
            self.container.subscribe(handler)
            return handler
        return decorator

    def get_next_event(self, *event_types, **kwargs):
        kwargs['once'] = True
        result = AsyncResult()

        @self.subscribe(*event_types, **kwargs)
        def handler(interface, event):
            result.set(event.body)

        return AsyncResultWrapper(self.container, handler, result)


class DefaultInterface(Interface):
    register_with_coordinator = False

    @rpc()
    def ping(self, payload=None):
        return payload

    @rpc()
    def status(self):
        return {
            'endpoint': self.container.endpoint,
            'identity': self.container.identity,
        }

    @rpc()
    def inspect(self):
        """
        Returns a description of all available rpc methods of this service
        """
        methods = []
        for interface_name, interface in list(self.container.installed_interfaces.items()):
            for name, func in six.iteritems(interface.methods):
                methods.append({
                    'name': '%s.%s' % (interface_name, name),
                    'params': list(func.args.args),
                    'help': textwrap.dedent(func.__doc__ or '').strip(),
                })
        return {
            'methods': methods,
        }

    @rpc()
    def get_metrics(self):
        return list(self.container.metrics_aggregator.get_metrics())



import gc

import resource
import gevent

from . import metrics


RUSAGE_ATTRS = (
    'utime', 'stime',
    'maxrss', 'ixrss', 'idrss', 'isrss',
    'minflt', 'majflt', 'nswap',
    'inblock', 'oublock',
    'msgsnd', 'msgrcv',
    'nsignals', 'nvcsw', 'nivcsw',
)


class GlobalMetrics(object):
    def __call__(self):
        return iter(self)


class RUsageMetrics(GlobalMetrics):
    def __init__(self, name='rusage'):
        self.attr_map = [('ru_{}'.format(attr), '{}.{}'.format(name, attr)) for attr in RUSAGE_ATTRS]

    def __iter__(self):
        ru = resource.getrusage(resource.RUSAGE_SELF)
        for ru_attr, series_name in self.attr_map:
            yield metrics.RawMetric(series_name, getattr(ru, ru_attr))


class GarbageCollectionMetrics(GlobalMetrics):
    def __init__(self, name='gc'):
        self.name = name

    def __iter__(self):
        yield metrics.RawMetric('{}.garbage'.format(self.name), len(gc.garbage))
        for i, count in enumerate(gc.get_count()):
            yield metrics.RawMetric('{}.count{}'.format(self.name, i), count)


class GeventMetrics(GlobalMetrics):
    def __init__(self, name='gevent'):
        self.name = name

    def __iter__(self):
        hub = gevent.get_hub()
        threadpool, loop = hub.threadpool, hub.loop
        yield metrics.RawMetric('gevent.threadpool.size', threadpool.size)
        yield metrics.RawMetric('gevent.threadpool.maxsize', threadpool.maxsize)
        yield metrics.RawMetric('gevent.active', loop.activecnt)
        yield metrics.RawMetric('gevent.pending', loop.pendingcnt)
        yield metrics.RawMetric('gevent.depth', loop.depth)



import inspect

import lymph
from lymph.core.decorators import RPCBase

import sphinx
from sphinx.ext.autodoc import MethodDocumenter, ClassDocumenter, setup as autodoc_setup


class RPCMethodDocumenter(MethodDocumenter):
    """Documenter for RPC methods."""

    # Priority must be higher than AttributeDocumenter since the data
    # descriptor RPC decorator should treated in documentation as a
    # method and not an attribute.
    priority = 11

    @classmethod
    def can_document_member(cls, member, membername, isattr, parent):
        return isinstance(member, RPCBase)

    def format_args(self):
        """Override argument extraction by getting it from RPC decorator."""
        args = inspect.formatargspec(*self.object.args)
        return args.replace('\\', '\\\\')

    def generate(self, *args, **kwargs):
        super(RPCMethodDocumenter, self).generate(*args, **kwargs)
        # If RPC decorator define exception to raise (e.g. _RPCDecorator),
        # include this laters in the documentation of the method.
        raises = getattr(self.object, 'raises', ())
        for ex in raises:
            self.add_line(u':raises %s: ' % ex.__name__, '<autodoc>')


class RPCInterfaceDocumenter(ClassDocumenter):
    """Documenter for RPC Lymph Interfaces."""

    @classmethod
    def can_document_member(cls, member, membername, isattr, parent):
        document = super(RPCInterfaceDocumenter, cls).can_document_member(member, membername, isattr, parent)
        return document and issubclass(member, lymph.Interface)

    def format_args(self):
        """Return empty since we don't want to document Interface __init__ argument
        since they are irrelevant for service usage.
        """
        return ''

    def filter_members(self, members, want_all):
        """Filter interface attribute to only document RPC methods."""
        members = super(RPCInterfaceDocumenter, self).filter_members(members, want_all)
        ret = []
        for name, obj, isattr in members:
            if isinstance(obj, RPCBase):
                ret.append((name, obj, isattr))
        return ret


def setup(app):
    autodoc_setup(app)

    app.add_autodocumenter(RPCMethodDocumenter)
    app.add_autodocumenter(RPCInterfaceDocumenter)

    return sphinx.__version__



# -*- coding: utf-8 -*-
import sys
import telnetlib
import logging

from lymph.client import Client
from lymph.cli.base import Command


logger = logging.getLogger(__name__)


class ShellCommand(Command):
    """
    Usage: lymph shell [options]

    Options:
      --remote=<name:identity-prefix>     Service instance name and identity.
      --guess-external-ip, -g             Guess the public facing IP of this machine and
                                          use it instead of the provided address.

    {COMMON_OPTIONS}

    Example:

        $ lymph shell --remote=echo:38428b071a6 --guess-external-ip

    """

    short_description = 'Open an interactive Python shell locally or remotely.'

    def run(self, **kwargs):
        self.client = Client.from_config(self.config)

        service_fullname = self.args.get('--remote')
        if service_fullname:
            return self._open_remote_shell(service_fullname)
        else:
            return self._open_local_shell()

    def get_imported_objects(self):
        return {'client': self.client, 'config': self.config}

    def _open_local_shell(self):
        imported_objects = self.get_imported_objects()
        try:
            import IPython
        except ImportError:
            IPython = None

        if IPython:
            IPython.start_ipython(
                argv=[],
                user_ns=imported_objects,
                banner1='Welcome to the lymph shell'
            )
        else:
            import code
            code.interact(local=imported_objects)

    def _open_remote_shell(self, service_fullname):
        backdoor_endpoint = self._get_backdoor_endpoint(service_fullname)
        if not backdoor_endpoint:
            return "No backdoor setup for %s" % service_fullname

        host, port = backdoor_endpoint.split(':')

        self._open_telnet(host, port)

    def _get_backdoor_endpoint(self, service_fullname):
        try:
            name, identity_prefix = service_fullname.split(':')
        except ValueError:
            sys.exit("Malformed argument it should be in the format 'name:identity'")
        service = self.client.container.lookup(name)
        instance = service.get_instance(identity_prefix)
        if instance is None:
            sys.exit('Unkown instance %s' % service_fullname)
        return instance.backdoor_endpoint

    def _open_telnet(self, host, port):
        telnet = telnetlib.Telnet()
        telnet.open(host, port)

        telnet.interact()



import abc
import datetime
import decimal
import functools
import json
import msgpack
import six
import uuid

from lymph.utils import Undefined, UndefinedType


@six.add_metaclass(abc.ABCMeta)
class ExtensionTypeSerializer(object):
    @abc.abstractmethod
    def serialize(self, obj):
        raise NotImplementedError

    @abc.abstractmethod
    def deserialize(self, obj):
        raise NotImplementedError


class DatetimeSerializer(ExtensionTypeSerializer):
    format = '%Y-%m-%dT%H:%M:%SZ'

    def serialize(self, obj):
        return obj.strftime(self.format)

    def deserialize(self, obj):
        return datetime.datetime.strptime(obj, self.format)


class DateSerializer(DatetimeSerializer):
    format = '%Y-%m-%d'

    def deserialize(self, obj):
        return super(DateSerializer, self).deserialize(obj).date()


class TimeSerializer(DatetimeSerializer):
    format = '%H:%M:%SZ'

    def deserialize(self, obj):
        return super(TimeSerializer, self).deserialize(obj).time()


class StrSerializer(ExtensionTypeSerializer):
    def __init__(self, factory):
        self.factory = factory

    def serialize(self, obj):
        return str(obj)

    def deserialize(self, obj):
        return self.factory(obj)


class SetSerializer(ExtensionTypeSerializer):
    def serialize(self, obj):
        return list(obj)

    def deserialize(self, obj):
        return set(obj)


class UndefinedSerializer(ExtensionTypeSerializer):
    def serialize(self, obj):
        return ''

    def deserialize(self, obj):
        return Undefined


_extension_type_serializers = {
    'datetime': DatetimeSerializer(),
    'date': DateSerializer(),
    'time': TimeSerializer(),
    'Decimal': StrSerializer(decimal.Decimal),
    'UUID': StrSerializer(uuid.UUID),
    'set': SetSerializer(),
    'UndefinedType': UndefinedSerializer(),
}


class BaseSerializer(object):
    def __init__(self, dumps=None, loads=None, load=None, dump=None):
        self._dumps = dumps
        self._loads = loads
        self._load = load
        self._dump = dump

    def dump_object(self, obj):
        obj_type = type(obj)
        serializer = _extension_type_serializers.get(obj_type.__name__)
        if serializer:
            obj = {
                '__type__': obj_type.__name__,
                '_': serializer.serialize(obj),
            }
        return obj

    def load_object(self, obj):
        obj_type = obj.get('__type__')
        if obj_type:
            serializer = _extension_type_serializers.get(obj_type)
            return serializer.deserialize(obj['_'])
        return obj

    def dumps(self, obj):
        return self._dumps(obj, default=self.dump_object)

    def loads(self, s):
        return self._loads(s, object_hook=self.load_object)

    def dump(self, obj, f):
        return self._dump(obj, f, default=self.dump_object)

    def load(self, f):
        return self._load(f, object_hook=self.load_object)


msgpack_serializer = BaseSerializer(
    dumps=functools.partial(msgpack.dumps, use_bin_type=True),
    loads=functools.partial(msgpack.loads, encoding='utf-8'),
    dump=functools.partial(msgpack.dump, use_bin_type=True),
    load=functools.partial(msgpack.load, encoding='utf-8'),
)

json_serializer = BaseSerializer(dumps=json.dumps, loads=json.loads, dump=json.dump, load=json.load)



import abc
import six

from lymph.core.components import Component


@six.add_metaclass(abc.ABCMeta)
class BaseEventSystem(Component):
    @classmethod
    def from_config(cls, config, **kwargs):
        return cls(**kwargs)

    def install(self, container):
        self.container = container

    def on_start(self):
        pass

    def on_stop(self, **kwargs):
        pass

    def subscribe(self, handler):
        raise NotImplementedError

    def unsubscribe(self, handler):
        raise NotImplementedError

    @abc.abstractmethod
    def emit(self, event, delay=0):
        raise NotImplementedError



import json

from lymph.client import Client
from lymph.cli.base import Command


class EmitCommand(Command):
    """
    Usage: lymph emit <event-type> [<body>] [options]

    {COMMON_OPTIONS}
    """

    short_description = 'Manually emits an event.'

    def run(self):
        event_type = self.args.get('<event-type>')
        body = json.loads(self.args.get('<body>'))

        client = Client.from_config(self.config)
        client.emit(event_type, body)




import gevent
import msgpack
import redis
import time

from lymph.core.interfaces import Interface
from lymph.core.decorators import rpc
from lymph.utils import make_id


class Scheduler(Interface):
    service_type = 'scheduler'
    schedule_key = 'schedule'

    def __init__(self, *args, **kwargs):
        super(Scheduler, self).__init__(*args, **kwargs)
        self.redis = redis.StrictRedis()

    def on_start(self):
        self.container.spawn(self.loop)

    @rpc()
    def schedule(self, eta, event_type, payload):
        self.redis.zadd(self.schedule_key, eta, msgpack.dumps({
            'id': make_id(),
            'event_type': event_type,
            'payload': payload,
        }))

    def loop(self):
        while True:
            pipe = self.redis.pipeline()
            now = int(time.time())
            pipe.zrangebyscore(self.schedule_key, 0, now)
            pipe.zremrangebyscore(self.schedule_key, 0, now)
            events, n = pipe.execute()
            for event in events:
                event = msgpack.loads(event, encoding='utf-8')
                self.emit(event['event_type'], event['payload'])
            gevent.sleep(1)



from __future__ import print_function

import logging
import collections

import zmq.green as zmq

from lymph.cli.base import Command
from lymph.client import Client
from lymph.core import services
from lymph.utils.logging import get_loglevel


class RemoteTail(collections.Iterator):
    """Tail remotely a stream of services published messages.

    Instance of this class implement the iterator protocol, which return
    messages as published by the remote services that this instance is register
    to.

    """

    Entry = collections.namedtuple('Entry', 'topic instance msg')

    def __init__(self, ctx=None):
        if ctx is None:
            ctx = zmq.Context.instance()

        self._sock = ctx.socket(zmq.SUB)
        self._sock.setsockopt_string(zmq.SUBSCRIBE, u'')
        self._instances = {}

    @property
    def instances(self):
        return self._instances

    def _on_status_change(self, instance, action):
        """Connect to a given service instance."""
        if action == services.ADDED:
            self._connect(instance)
        elif action == services.REMOVED:
            self._disconnect(instance)

    def _connect(self, instance):
        self._sock.connect(instance.log_endpoint)
        self._instances[instance.log_endpoint] = instance

    def _disconnect(self, instance):
        self._sock.disconnect(instance.log_endpoint)
        del self._instances[instance.log_endpoint]

    def subscribe_service(self, service):
        """Subscribe to a service stream.

        This is done by iterating over all instances in this service and
        connecting to them, while keeping tabs over this service to be able
        to connect and disconnect as instances get added or removed.

        Return: True if subscription worked else false.

        """
        service.observe([services.ADDED, services.REMOVED], self._on_status_change)

        connected = False
        for instance in service:
            if instance.log_endpoint:
                self._connect(instance)
                connected = True
        return connected

    def next(self):
        """Return an instance of :class:`RemoteTail.Entry`."""
        topic, endpoint, msg = self._sock.recv_multipart()
        return self.Entry(topic, self._instances[endpoint], msg)

    __next__ = next  # For python3.


class TailCommand(Command):
    """
    Usage: lymph tail [options] [--level=<level> | -l <level>] <address>...

    Description:
        Shows the log output of <address> on stderr

    Tail Options:
      --level=<level>, -l <level>  Log level to subscribe to [default: INFO]

    {COMMON_OPTIONS}
    """

    short_description = 'Stream the logs of one or more services.'

    def run(self):
        client = Client.from_config(self.config)
        tail = RemoteTail()

        for address in self.args['<address>']:
            connected = tail.subscribe_service(client.container.lookup(address))
            if not connected:
                print("Couldn't connect to log endpoint of '%s'" % address)

        if not tail.instances:
            return 1

        level = get_loglevel(self.args['--level'])
        logger = logging.getLogger('lymph-tail-cli')
        logger.setLevel(level)

        console = logging.StreamHandler()
        console.setLevel(level)
        console.setFormatter(logging.Formatter('[%(service_type)s][%(identity)s] [%(levelname)s] %(message)s'))

        logger.addHandler(console)

        try:
            for topic, instance, msg in tail:
                level = getattr(logging, topic)
                extra = {
                    'identity': instance.identity[:10],
                    'service_type': instance.endpoint,
                }
                logger.log(level, msg, extra=extra)
        except KeyboardInterrupt:
            pass



from __future__ import absolute_import, division, print_function

import collections
import importlib
import inspect
import gc
import gevent
import math
import os
import re
import sys
import threading
import traceback
import uuid


class UndefinedType(object):
    def __repr__(self):
        return "Undefined"

    def __bool__(self):
        return False

    def __nonzero__(self):
        return False


Undefined = UndefinedType()


def import_object(module_name, object_path=None):
    if not object_path:
        if ':' not in module_name:
            raise ValueError("cannot import object %r" % module_name)
        module_name, object_path = module_name.split(':')
    mod = importlib.import_module(module_name)
    obj = mod
    for objname in object_path.split('.'):
        obj = getattr(obj, objname)
    return obj


def make_id():
    return uuid.uuid4().hex


_sqrt2 = math.sqrt(2)


class Accumulator(object):
    def __init__(self):
        self.n = 0
        self.sum = 0
        self.square_sum = 0
        self._mean = None
        self._stddev = None

    def add(self, value):
        self.n += 1
        self.sum += value
        self.square_sum += value * value
        self._mean = None
        self._stddev = None

    def remove(self, value):
        self.n -= 1
        self.sum -= value
        self.square_sum -= value * value
        self._mean = None
        self._stddev = None

    @property
    def mean(self):
        if not self.n:
            return 0.
        if self._mean is None:
            self._mean = self.sum / self.n
        return self._mean

    @property
    def stddev(self):
        if not self.n:
            return 0.
        if self._stddev is None:
            mean = self.mean
            self._stddev = math.sqrt(self.square_sum / self.n - mean * mean)
        return self._stddev

    @property
    def stats(self):
        return {'mean': self.mean, 'stddev': self.stddev, 'n': self.n}


class SampleWindow(Accumulator):
    def __init__(self, n=100, factor=1):
        super(SampleWindow, self).__init__()
        self.size = n
        self.factor = factor
        self.values = collections.deque([])
        self.total = Accumulator()

    def __len__(self):
        return len(self.values)

    def is_full(self):
        return len(self.values) == self.size

    def add(self, value):
        value = value * self.factor
        super(SampleWindow, self).add(value)
        self.total.add(value)
        if self.is_full():
            self.remove(self.values.popleft())
        self.values.append(value)

    def p(self, value):
        """
        returns the probability for samples greater than `value` given a normal
        distribution with mean and standard deviation derived from this window.
        """
        if self.stddev == 0:
            return 1. if value == self.mean else 0.
        return 1 - math.erf(abs(value * self.factor - self.mean) / (self.stddev * _sqrt2))


def get_greenlets():
    for object in gc.get_objects():
        if isinstance(object, gevent.Greenlet):
            yield object


def get_greenlets_frames():
    for greenlet in get_greenlets():
        yield str(greenlet), greenlet.gr_frame


def get_threads_frames():
    threads = {thread.ident: thread.name for thread in threading.enumerate()}
    for ident, frame in sys._current_frames().items():
        name = threads.get(ident)
        if name:
            yield '%s:%s' % (ident, name), frame


def format_stack(frame):
    tb = traceback.format_stack(frame)
    return ''.join(tb)


def dump_stacks(output=print):
    output('PID: %s' % os.getpid())
    output('Threads')
    for i, (name, frame) in enumerate(get_threads_frames()):
        output('Thread #%d: %s' % (i, name))
        output(format_stack(frame))
    output('Greenlets')
    for i, (name, frame) in enumerate(get_greenlets_frames()):
        output('Greenlet #%d: %s' % (i, name))
        output(format_stack(frame))



# -*- coding: utf-8 -*-
from __future__ import division, unicode_literals

import gevent
import math
import os
import time
import logging

from lymph.utils import SampleWindow
from lymph.exceptions import RpcError

logger = logging.getLogger(__name__)

UNKNOWN = 'unknown'
RESPONSIVE = 'responsive'
UNRESPONSIVE = 'unresponsive'
CLOSED = 'closed'
IDLE = 'idle'


class Connection(object):
    def __init__(self, server, endpoint, heartbeat_interval=1, timeout=3, idle_timeout=10, unresponsive_disconnect=30, idle_disconnect=60):
        assert heartbeat_interval < timeout < idle_timeout
        self.server = server
        self.endpoint = endpoint
        self.timeout = timeout
        self.heartbeat_interval = heartbeat_interval
        self.idle_timeout = idle_timeout
        self.unresponsive_disconnect = unresponsive_disconnect
        self.idle_disconnect = idle_disconnect

        now = time.monotonic()
        self.last_seen = 0
        self.idle_since = 0
        self.last_message = now
        self.created_at = now
        self.heartbeat_samples = SampleWindow(100, factor=1000)  # milliseconds
        self.explicit_heartbeat_count = 0
        self.status = UNKNOWN

        self.received_message_count = 0
        self.sent_message_count = 0

        self.heartbeat_loop_greenlet = self.server.spawn(self.heartbeat_loop)
        self.live_check_loop_greenlet = self.server.spawn(self.live_check_loop)

        self.pid = os.getpid()

    def __str__(self):
        return "connection to=%s last_seen=%s" % (self.endpoint, self._dt())

    def _dt(self):
        return time.monotonic() - self.last_seen

    @property
    def phi(self):
        p = self.heartbeat_samples.p(self._dt())
        if p == 0:
            return float('inf')
        return -math.log10(p)

    def set_status(self, status):
        self.status = status

    def heartbeat_loop(self):
        while True:
            start = time.monotonic()
            channel = self.server.ping(self.endpoint)
            error = False
            try:
                channel.get(timeout=self.heartbeat_interval)
            except RpcError as e:
                logger.debug('hearbeat error on %s: %r', self, e)
                error = True
            took = time.monotonic() - start
            if not error:
                self.heartbeat_samples.add(took)
                self.explicit_heartbeat_count += 1
            gevent.sleep(max(0, self.heartbeat_interval - took))

    def live_check_loop(self):
        while True:
            self.update_status()
            self.log_stats()
            gevent.sleep(1)

    def update_status(self):
        if self.last_seen:
            now = time.monotonic()
            if now - self.last_seen >= self.timeout:
                self.set_status(UNRESPONSIVE)
            elif now - self.last_message >= self.idle_timeout:
                self.set_status(IDLE)
                self.idle_since = now
            else:
                self.set_status(RESPONSIVE)

    def log_stats(self):
        roundtrip_stats = 'window (mean rtt={mean:.1f} ms; stddev rtt={stddev:.1f})'.format(**self.heartbeat_samples.stats)
        roundtrip_total_stats = 'total (mean rtt={mean:.1f} ms; stddev rtt={stddev:.1f})'.format(**self.heartbeat_samples.total.stats)
        logger.debug("pid=%s; endpoint=%s; %s; %s; phi=%.3f; ping/s=%.2f; status=%s" % (
            self.pid,
            self.endpoint,
            roundtrip_stats,
            roundtrip_total_stats,
            self.phi,
            self.explicit_heartbeat_count / max(1, time.monotonic() - self.created_at),
            self.status,
        ))

    def close(self):
        if self.status == CLOSED:
            return
        self.status = CLOSED
        self.heartbeat_loop_greenlet.kill()
        self.live_check_loop_greenlet.kill()
        self.server.disconnect(self.endpoint)

    def on_recv(self, msg):
        now = time.monotonic()
        self.last_seen = now
        if not msg.is_idle_chatter():
            self.last_message = now
        self.received_message_count += 1

    def on_send(self, msg):
        if not msg.is_idle_chatter():
            self.last_message = time.monotonic()
        self.sent_message_count += 1

    def is_alive(self):
        return self.status in (RESPONSIVE, IDLE, UNKNOWN)

    def stats(self):
        # FIXME: rtt and phi should be recorded as summary/histogram for all connections
        return {
            'endpoint': self.endpoint,
            'rtt': self.heartbeat_samples.stats,
            'phi': self.phi,
            'status': self.status,
            'sent': self.sent_message_count,
            'received': self.received_message_count,
        }



import re
import unittest

import mock

import lymph
from lymph.testing import RPCServiceTestCase
from lymph.testing.mock_helpers import MockMixins, RpcMockTestCase, EventMockTestCase


class DummyTestCase(unittest.TestCase, MockMixins):

    def runTest(self):
        pass


class RPCMockHelperTests(unittest.TestCase):

    def setUp(self):
        self.dummy_case = DummyTestCase()

    def test_single_call_match(self):
        self.dummy_case._assert_equal_calls(
            [mock.call('func', 1, foo='bar')],
            [mock.call('func', 1, foo='bar')]
        )

    def test_single_call_different_name(self):
        with self.assertRaisesRegexp(AssertionError, "function #0 name doesn't match, expected 'func2' actual 'func1'"):
            self.dummy_case._assert_equal_calls(
                [mock.call('func1', 1)],
                [mock.call('func2', 1)]
            )

    def test_single_call_args_mismatch(self):
        with self.assertRaisesRegexp(AssertionError, re.compile("function #0 argument #0 doesn't match.*", re.DOTALL)):
            self.dummy_case._assert_equal_calls(
                [mock.call('func', 1, foo='bar')],
                [mock.call('func', 102, foo='bar')]
            )

    def test_single_call_keyword_value_mismatch(self):
        with self.assertRaisesRegexp(AssertionError, re.compile("function #0 keyword argument 'foo' doesn't match.*", re.DOTALL)):
            self.dummy_case._assert_equal_calls(
                [mock.call('func', 1, foo='foobar')],
                [mock.call('func', 1, foo='bar')]
            )

    def test_single_call_keyword_name_mismatch(self):
        with self.assertRaisesRegexp(AssertionError, "function #0 keyword arguments doesn't match, expected \['something'\] actual \['foo'\]"):
            self.dummy_case._assert_equal_calls(
                [mock.call('func', 1, foo='bar')],
                [mock.call('func', 1, something='bar')]
            )

    def test_single_call_keyword_different_count(self):
        with self.assertRaisesRegexp(AssertionError, "function #0 keyword arguments doesn't match, expected \['bar', 'foo'\] actual \['foo'\]"):
            self.dummy_case._assert_equal_calls(
                [mock.call('func', 1, foo='bar')],
                [mock.call('func', 1, foo='bar', bar='taz')]
            )

    def test_single_call_argument_different_count(self):
        with self.assertRaisesRegexp(AssertionError, "function #0 arguments count doesn't match, expected 1 actual 2"):
            self.dummy_case._assert_equal_calls(
                [mock.call('func', 1, 2)],
                [mock.call('func', 1)]
            )

    def test_multiple_call_match(self):
        self.dummy_case._assert_equal_calls(
            [
                mock.call('func', 1, foo='foo'),
                mock.call('func', 2, foo='bar'),
                mock.call('func', 3, foo='foobar')
            ],
            [
                mock.call('func', 1, foo='foo'),
                mock.call('func', 2, foo='bar'),
                mock.call('func', 3, foo='foobar')
            ],
        )

    def test_assert_equal_any_call_success(self):
        self.dummy_case._assert_equal_any_calls(
            [
                mock.call('func1', 1, foo='foo'),
                mock.call('func2', 2, foo='bar'),
                mock.call('func3', 3, foo='foobar')
            ],
            [
                mock.call('func2', 2, foo='bar'),
                mock.call('func3', 3, foo='foobar')
            ],
        )

    def test_assert_equal_any_call_success_with_no_expect(self):
        self.dummy_case._assert_equal_any_calls(
            [
                mock.call('func1', 1, foo='foo'),
                mock.call('func2', 2, foo='bar'),
                mock.call('func3', 3, foo='foobar')
            ],
            [],
        )

    def test_assert_equal_any_call_fail_with_possible_matches(self):
        with self.assertRaisesRegexp(AssertionError, "Call 'call\('func10', 3, foo='foobar'\)' wasn't found."):
            self.dummy_case._assert_equal_any_calls(
                [
                    mock.call('func1', 1, foo='foo'),
                    mock.call('func2', 2, foo='bar'),
                    mock.call('func3', 3, foo='foobar')
                ],
                [
                    mock.call('func10', 3, foo='foobar')
                ],
            )

    def test_assert_equal_any_call_fail_with_no_match(self):
        with self.assertRaisesRegexp(AssertionError, re.compile("Call 'call\('func2', 3, foo='foo'\)' wasn't found. Maybe you want:.*?", re.DOTALL)):
            self.dummy_case._assert_equal_any_calls(
                [
                    mock.call('func2', 1, foo='foo'),
                    mock.call('func2', 2, foo='bar'),
                    mock.call('func2', 3, foo='foobar')
                ],
                [
                    mock.call('func2', 3, foo='foo')
                ],
            )

    def test_assert_equal_any_call_fail_with_no_match_better_message(self):
        with self.assertRaisesRegexp(AssertionError, re.compile("function #0 keyword argument 'foo' doesn't match*?", re.DOTALL)):
            self.dummy_case._assert_equal_any_calls(
                [
                    mock.call('func1', 2, foo='bar'),
                    mock.call('func2', 3, foo='foobar')
                ],
                [
                    mock.call('func2', 3, foo='foo')
                ],
            )

    def test_assert_equal_any_call_fail_with_wrong_order(self):
        with self.assertRaisesRegexp(AssertionError, re.compile("Call 'call\('func2', 2, foo='bar'\)' wasn't found.", re.DOTALL)):
            self.dummy_case._assert_equal_any_calls(
                [
                    mock.call('func1', 1, foo='foo'),
                    mock.call('func2', 2, foo='bar'),
                    mock.call('func3', 3, foo='foobar')
                ],
                [
                    mock.call('func3', 3, foo='foobar'),
                    mock.call('func2', 2, foo='bar')
                ],
            )


class StringService(lymph.Interface):

    @lymph.rpc()
    def upper(self, text):
        self.emit('str.uppered', {'text': text})
        return text.upper()

    @lymph.rpc()
    def lower(self, text):
        self.emit('str.lowered', {'text': text})
        return text.lower()


class MetaRPCUpperTestCase(RPCServiceTestCase, RpcMockTestCase, EventMockTestCase):

    service_class = StringService
    service_name = 'str'

    def setUp(self):
        super(MetaRPCUpperTestCase, self).setUp()
        self.setup_rpc_mocks({
            'str.upper': 'HELLO WORLD',
            'str.lower': 'hello world'
        })

    def test_meta_rpc(self):
        response = self.client.upper(text='hello world')

        self.assertEqual(response, 'HELLO WORLD')

        self.assert_rpc_calls(
            mock.call('str.upper', text='hello world')
        )

    def test_meta_events(self):
        self.delete_rpc_mock('str.upper')

        response = self.client.upper(text='hello world')

        self.assertEqual(response, 'HELLO WORLD')

        self.assert_rpc_calls(
            mock.call('str.upper', text='hello world')
        )

        self.assert_events_emitted(
            mock.call('str.uppered', {'text': 'hello world'})
        )

    def test_meta_update_rpc(self):
        self.update_rpc_mock('str.upper', 'FOOBAR')

        response = self.client.upper(text='hello world')

        self.assertEqual(response, 'FOOBAR')

        self.assert_rpc_calls(
            mock.call('str.upper', text='hello world')
        )

    def test_meta_multiple_rpc(self):
        response = self.client.upper(text='hello world')
        self.assertEqual(response, 'HELLO WORLD')

        response = self.client.lower(text='HELLO WORLD')
        self.assertEqual(response, 'hello world')

        self.assert_rpc_calls(
            mock.call('str.upper', text='hello world'),
            mock.call('str.lower', text='HELLO WORLD'),
        )



import gevent
import gevent.queue

from lymph.exceptions import Timeout, Nack, RemoteError
from lymph.core.messages import Message


class Channel(object):
    def __init__(self, request, server):
        self.request = request
        self.server = server


class RequestChannel(Channel):
    def __init__(self, request, server):
        super(RequestChannel, self).__init__(request, server)
        self.queue = gevent.queue.Queue()

    def recv(self, msg):
        self.queue.put(msg)

    def get(self, timeout=1):
        try:
            msg = self.queue.get(timeout=timeout)
            if msg.type == Message.NACK:
                raise Nack(self.request)
            elif msg.type == Message.ERROR:
                raise RemoteError.from_reply(self.request, msg)
            return msg
        except gevent.queue.Empty:
            raise Timeout(self.request)
        finally:
            self.close()

    def close(self):
        del self.server.channels[self.request.id]


class ReplyChannel(Channel):
    def __init__(self, request, server):
        super(ReplyChannel, self).__init__(request, server)
        self._sent_reply = False

    def reply(self, body):
        self.server.send_reply(self.request, body)
        self._sent_reply = True

    def ack(self, unless_reply_sent=False):
        if unless_reply_sent and self._sent_reply:
            return
        self.server.send_reply(self.request, None, msg_type=Message.ACK)
        self._sent_reply = True

    def nack(self, unless_reply_sent=False):
        if unless_reply_sent and self._sent_reply:
            return
        self.server.send_reply(self.request, None, msg_type=Message.NACK)
        self._sent_reply = True

    def error(self, **body):
        self.server.send_reply(self.request, body, msg_type=Message.ERROR)

    def close(self):
        pass



from __future__ import division

from six.moves import range
from unittest import TestCase

from lymph.utils import Accumulator


class AccumulatorTests(TestCase):
    def test_accumulator(self):
        acc = Accumulator()
        for i in range(1, 6):
            acc.add(i / 15.)
        self.assertEqual(acc.n, 5)
        self.assertEqual(acc.sum, 1)
        self.assertEqual(acc.mean, 0.2)
        self.assertEqual(acc.stddev, 0.09428090415820631)



import logging
import sys

from werkzeug.contrib.wrappers import DynamicCharsetRequestMixin
from werkzeug.wrappers import Request as BaseRequest, Response
from werkzeug.exceptions import HTTPException

from lymph.core.interfaces import Interface
from lymph.core import trace
from lymph.utils.logging import setup_logger
from lymph.exceptions import SocketNotCreated
from lymph.utils import sockets
from lymph.core.trace import Group
from lymph.web.wsgi_server import LymphWSGIServer
from lymph.web.routing import HandledRule


logger = logging.getLogger(__name__)


class Request(DynamicCharsetRequestMixin, BaseRequest):
    default_charset = 'utf-8'

    @property
    def full_path(self):
        full_path = super(Request, self).full_path
        return full_path.rstrip('?')


class WebServiceInterface(Interface):
    default_http_port = None

    def __init__(self, *args, **kwargs):
        super(WebServiceInterface, self).__init__(*args, **kwargs)
        self.application = Request.application(self.dispatch_request)
        if self.container.debug:
            from werkzeug.debug import DebuggedApplication
            self.application = DebuggedApplication(self.application, evalex=True)
        self.wsgi_server = None

    def __call__(self, *args, **kwargs):
        # Make the object itself a WSGI app
        return self.application(*args, **kwargs)

    def apply_config(self, config):
        super(WebServiceInterface, self).apply_config(config)
        self.http_port = config.get('port', self.default_http_port)
        self.pool_size = config.get('wsgi_pool_size')
        if self.http_port is None:
            self.http_port = sockets.get_unused_port()

    def get_description(self):
        description = super(WebServiceInterface, self).get_description()
        description['http_port'] = self.http_port
        return description

    def on_start(self):
        super(WebServiceInterface, self).on_start()
        setup_logger('werkzeug')
        setup_logger('gevent')
        try:
            socket_fd = self.container.get_shared_socket_fd(self.http_port)
        except SocketNotCreated:
            logger.warning("socket for port %s wasn't created by node, binding from instance instead", self.http_port)
            address = '%s:%s' % (self.container.server.ip, self.http_port)
            self.http_socket = sockets.create_socket(address)
        else:
            self.http_socket = sockets.create_socket('fd://%s' % socket_fd)
        self.wsgi_server = LymphWSGIServer(self.http_socket, self.application, spawn=Group(self.pool_size))
        self.wsgi_server.start()

    def on_stop(self, **kwargs):
        self.wsgi_server.stop()
        super(WebServiceInterface, self).on_stop()

    def dispatch_request(self, request):
        trace.set_id()
        logger.info('%s %s', request.method, request.path)
        urls = self.url_map.bind_to_environ(request.environ)
        request.urls = urls
        try:
            rule, kwargs = request.urls.match(return_rule=True)
            self.container.http_request_hook(request, rule, kwargs)
            handler = self.get_handler(request, rule)
            if hasattr(handler, "dispatch"):
                response = handler.dispatch(kwargs)
            else:
                response = handler(request, **kwargs)
        except HTTPException as e:
            response = e.get_response(request.environ)
        except Exception as e:
            if hasattr(e, 'to_http_response'):
                response = e.to_http_response(request.environ)
            else:
                if not self.container.debug:
                    logger.exception('uncaught exception')
                exc_info = sys.exc_info()
                extra_info = {
                    'url': request.url,
                    'trace_id': trace.get_id(),
                    'interface': self.__class__.__name__,
                }
                try:
                    self.container.error_hook(exc_info, extra=extra_info)
                except:
                    logger.exception('error hook failure')
                finally:
                    del exc_info
                if self.container.debug:
                    raise
                response = Response('', status=500)
        return response

    def get_handler(self, request, rule):
        if callable(rule.endpoint):
            handler = rule.endpoint(self, request)
        elif isinstance(rule, HandledRule):
            handler = rule.handler(self, request)
        else:
            handler = getattr(self, rule.endpoint)
        return handler

    def get_wsgi_application(self):
        return self.application



from __future__ import absolute_import
from datetime import (date, datetime)
import logging
import six
import uuid


logger = logging.getLogger(__name__)


class EventIndex(object):
    def __init__(self, es, index_name='events'):
        self.es = es
        self.index_name = index_name

    def prepare_object(self, data):
        return dict(self.prepare_value(key, value)
                    for key, value in six.iteritems(data))

    def prepare_value(self, key, value):
        if isinstance(value, bool):
            type_prefix = 'b'
        elif isinstance(value, six.integer_types):
            type_prefix = 'i'
        elif isinstance(value, six.string_types):
            type_prefix = 's'
        elif isinstance(value, float):
            type_prefix = 'f'
        elif isinstance(value, dict):
            type_prefix = 'o'
            value = self.prepare_object(value)
        elif isinstance(value, list):
            type_prefix = 'l'
        elif isinstance(value, (datetime, date)):
            type_prefix = 'd'
        elif isinstance(value, uuid.UUID):
            type_prefix = 'u'
            value = value.hex
        else:
            raise TypeError('cannot index values of type %s' % type(value))
        return ('%s_%s' % (type_prefix, key)), value

    def index(self, event, index_name=None):
        event_id = uuid.uuid4().hex
        body = self.prepare_object(event.body)
        body.update({
            'type': event.evt_type,
            'source': event.source,
            'logged_at': datetime.utcnow(),
        })
        self.es.index(
            index=index_name or self.index_name,
            doc_type='event',
            id=event_id,
            body=body,
        )


class DatedEventIndex(EventIndex):
    def create_index_alias(self):
        if self.es.indices.exists_alias(self.index_name):
            logger.info('index alias already exists')
        self.es.indices.put_alias(
            index='%s-*' % self.index_name,
            name=self.index_name,
        )

    def get_index_name(self, dt):
        return '%s-%s' % (self.index_name, dt.strftime('%Y.%m.%d'))

    def index(self, event, index_name=None):
        index_name = self.get_index_name(datetime.now())
        super(DatedEventIndex, self).index(event, index_name)




import json
import logging
import os
import sys
import socket

import six

from lymph.exceptions import RegistrationFailure, SocketNotCreated
from lymph.core.components import Componentized
from lymph.core.events import Event
from lymph.core.monitoring import metrics
from lymph.core.monitoring.pusher import MonitorPusher
from lymph.core.monitoring.aggregator import Aggregator
from lymph.core.services import ServiceInstance, Service
from lymph.core.rpc import ZmqRPCServer
from lymph.core.interfaces import DefaultInterface
from lymph.core.plugins import Hook
from lymph.core import trace


logger = logging.getLogger(__name__)


def create_container(config):
    if 'registry' in config:
        logger.warning('global `registry` configuration is deprecated. please use `container.registry` instead.')
        config.set('container.registry', config.get_raw('registry'))
    event_system = config.create_instance('event_system')
    container = config.create_instance(
        'container',
        default_class='lymph.core.container:ServiceContainer',
        events=event_system,
    )
    return container


class ServiceContainer(Componentized):
    def __init__(self, rpc=None, registry=None, events=None, log_endpoint=None, service_name=None, debug=False, monitor_endpoint=None, pool=None):
        if pool is None:
            pool = trace.Group()
        super(ServiceContainer, self).__init__(error_hook=Hook('error_hook'), pool=pool)
        self.log_endpoint = log_endpoint
        self.backdoor_endpoint = None
        self.service_name = service_name
        self.fqdn = socket.getfqdn()

        self.http_request_hook = Hook('http_request_hook')

        self.server = rpc
        self.service_registry = registry
        self.event_system = events

        self.installed_interfaces = {}
        self.installed_plugins = []

        self.debug = debug
        self.monitor_endpoint = monitor_endpoint

        self.metrics_aggregator = Aggregator(self._get_metrics, service=self.service_name, host=self.fqdn)

        if self.service_registry:
            self.add_component(self.service_registry)

        if self.event_system:
            self.add_component(self.event_system)
            self.event_system.install(self)

        self.monitor = self.install(MonitorPusher, aggregator=self.metrics_aggregator, endpoint=self.monitor_endpoint, interval=5)

        self.add_component(rpc)
        rpc.request_handler = self.handle_request

        self.install_interface(DefaultInterface, name='lymph')

    @classmethod
    def from_config(cls, config, **explicit_kwargs):
        kwargs = dict(config)
        kwargs.pop('class', None)
        kwargs.setdefault('monitor_endpoint', os.environ.get('LYMPH_MONITOR'))
        kwargs.setdefault('service_name', os.environ.get('LYMPH_SERVICE_NAME'))
        kwargs['registry'] = config.create_instance('registry')

        kwargs['rpc'] = config.create_instance('rpc', default_class=ZmqRPCServer, ip=kwargs.pop('ip', None), port=kwargs.pop('port', None))
        kwargs['pool'] = config.create_instance('pool', default_class='lymph.core.trace:Group')

        for key, value in six.iteritems(explicit_kwargs):
            if value is not None:
                kwargs[key] = value
        return cls(**kwargs)

    def excepthook(self, type, value, traceback):
        logger.log(logging.CRITICAL, 'Uncaught exception', exc_info=(type, value, traceback))
        self.error_hook((type, value, traceback))

    @property
    def endpoint(self):
        return self.server.endpoint

    @property
    def identity(self):
        return self.server.identity

    def install_interface(self, cls, **kwargs):
        interface = self.install(cls, **kwargs)
        self.installed_interfaces[interface.name] = interface
        for plugin in self.installed_plugins:
            plugin.on_interface_installation(interface)
        return interface

    def install_plugin(self, cls, **kwargs):
        plugin = self.install(cls, **kwargs)
        self.installed_plugins.append(plugin)
        return plugin

    def get_shared_socket_fd(self, port):
        fds = json.loads(os.environ.get('LYMPH_SHARED_SOCKET_FDS', '{}'))
        try:
            return fds[str(port)]
        except KeyError:
            raise SocketNotCreated

    @property
    def service_types(self):
        return self.installed_interfaces.keys()

    def subscribe(self, handler, **kwargs):
        return self.event_system.subscribe(handler, **kwargs)

    def unsubscribe(self, handler):
        self.event_system.unsubscribe(handler)

    def get_instance_description(self, interface):
        description = interface.get_description()
        description.update({
            'endpoint': self.endpoint,
            'identity': self.identity,
            'log_endpoint': self.log_endpoint,
            'backdoor_endpoint': self.backdoor_endpoint,
            'fqdn': self.fqdn,
            'ip': self.server.ip,
        })
        return description

    def start(self, register=True):
        logger.info('starting %s (%s) at %s (pid=%s)', self.service_name, ', '.join(self.service_types), self.endpoint, os.getpid())

        self.on_start()
        self.metrics_aggregator.add_tags(identity=self.identity)

        if register:
            for interface_name, interface in six.iteritems(self.installed_interfaces):
                if not interface.register_with_coordinator:
                    continue
                instance = ServiceInstance(**self.get_instance_description(interface))
                try:
                    self.service_registry.register(interface_name, instance)
                except RegistrationFailure:
                    logger.error("registration failed %s, %s", interface_name, interface)
                    self.stop()

    def stop(self, **kwargs):
        self.on_stop()
        self.pool.kill()

    def join(self):
        self.pool.join()

    def lookup(self, address):
        if '://' not in address:
            return self.service_registry.get(address)
        instance = ServiceInstance(address)
        return Service(address, instances=[instance])

    def discover(self):
        return self.service_registry.discover()

    def emit_event(self, event_type, payload, headers=None, **kwargs):
        headers = headers or {}
        headers.setdefault('trace_id', trace.get_id())
        event = Event(event_type, payload, source=self.identity, headers=headers)
        self.event_system.emit(event, **kwargs)

    def send_request(self, address, subject, body, headers=None):
        service = self.lookup(address)
        return self.server.send_request(service, subject, body, headers=headers)

    def handle_request(self, channel):
        interface_name, func_name = channel.request.subject.rsplit('.', 1)
        try:
            interface = self.installed_interfaces[interface_name]
        except KeyError:
            logger.warning('unsupported service type: %s', interface_name)
            channel.nack(True)
            return
        try:
            interface.handle_request(func_name, channel)
        except Exception:
            logger.exception('Request error:')
            exc_info = sys.exc_info()
            try:
                self.error_hook(exc_info, extra={
                    'service': self.service_name,
                    'interface': interface_name,
                    'func_name': func_name,
                    'trace_id': trace.get_id(),
                })
            finally:
                del exc_info
            try:
                channel.nack(True)
            except:
                logger.exception('failed to send automatic NACK')

    def _get_metrics(self):
        for metric in super(ServiceContainer, self)._get_metrics():
            yield metric
        yield metrics.RawMetric('greenlets.count', len(self.pool))




def setup_config(args):
    import os
    import sys
    import yaml

    from lymph.config import Configuration
    from lymph.utils.sockets import guess_external_ip

    vars_file = args.get('--vars') or os.environ.get('LYMPH_VARS')
    env_vars = Configuration()
    if vars_file:
        env_vars.load_file(vars_file)
        os.environ['LYMPH_VARS'] = vars_file

    config = Configuration({'container': {}}, env=os.environ, var=env_vars)

    if 'LYMPH_NODE_CONFIG' in os.environ:
        config.load_file(os.environ['LYMPH_NODE_CONFIG'], sections=['container', 'registry', 'event_system', 'plugins', 'dependencies'])

    config_file = args.get('--config') or '.lymph.yml'
    config.load_file(config_file)
    config.source = config_file

    config.setdefault('container.ip', os.environ.get('LYMPH_NODE_IP', '127.0.0.1'))
    ip = args.get('--ip')
    if args.get('--guess-external-ip'):
        if ip:
            sys.exit('Cannot combine --ip and --guess-external-ip')
        ip = guess_external_ip()
        if ip is None:
            sys.exit('Cannot guess external ip, aborting ...')
    if ip:
        config.set('container.ip', ip)

    port = args.get('--port')
    if port:
        config.set('container.port', port)

    return config


def setup_terminal(args, config):
    import blessings

    force_color = args.get('--color', False)
    if args.get('--no-color', False):
        if force_color:
            raise ValueError("cannot combine --color and --no-color")
        force_color = None
    return blessings.Terminal(force_styling=force_color)


def _excepthook(type, value, tb):
    import logging

    logger = logging.getLogger('lymph')
    logger.log(logging.CRITICAL, 'Uncaught exception', exc_info=(type, value, tb))


def main(argv=None):
    import lymph.monkey
    lymph.monkey.patch()

    import docopt
    import sys
    import logging

    from lymph import __version__ as VERSION
    from lymph.cli.help import HELP
    from lymph.cli.base import get_command_class
    from lymph.utils import logging as lymph_logging

    args = docopt.docopt(HELP, argv, version=VERSION, options_first=True)
    name = args.pop('<command>')
    argv = args.pop('<args>')
    try:
        command_cls = get_command_class(name)
    except KeyError:
        print("'%s' is not a valid lymph command. See 'lymph list' or 'lymph help'." % name)
        return 1
    command_args = docopt.docopt(command_cls.get_help(), [name] + argv)
    args.update(command_args)

    config = setup_config(args) if command_cls.needs_config else None

    if config:
        loglevel = args.get('--loglevel', 'ERROR')
        logfile = args.get('--logfile')

        lymph_logging.setup_logging(config, loglevel, logfile)
    else:
        logging.basicConfig()

    sys.excepthook = _excepthook

    terminal = setup_terminal(args, config)
    command = command_cls(args, config, terminal)
    return command.run()


if __name__ == '__main__':
    main()



from lymph.cli.base import Command, get_command_classes, get_command_class, format_docstring


HEADER = 'Usage: lymph [options] <command> [<args>...]'

HELP = HEADER + """
  lymph help             display help overview
  lymph help <command>   display command documentation
"""

TEMPLATE = HEADER + """

{COMMON_OPTIONS}

Commands:
%s
"""


def _format_help(name, description, indent='  ', spaces=13, min_spaces=2):
    r"""Format ``name`` + ``description`` in an unified format
    that can be used to print beautiful help messages.

    If the name is too long (length is greater than ``spaces - min_spaces``)
    than the name and description will appear in different lines.

    Example:

        >>> print(_format_help('foo', 'foobar'))
          foo          foobar
        >>> print(_format_help('foo', 'foobar', spaces=4))
          foo
              foobar
        >>> print('\n'.join([
        ...     _format_help('help', 'Print help message'),
        ...     _format_help('shell', 'Open an interactive Python shell.'),
        ...     _format_help('storage-migration', 'One big name for a command option'),
        ... ]))
        ...
          help         Print help message
          shell        Open an interactive Python shell.
          storage-migration
                       One big name for a command option


    """
    if spaces - len(name) < min_spaces:
        return '\n'.join([
            indent + name,
            indent + (' ' * spaces) + description
        ])
    else:
        return indent + name + (' ' * (spaces - len(name))) + description


class HelpCommand(Command):
    """
    Usage: lymph help [<command>]
    """

    short_description = 'Display help information about lymph.'
    needs_config = False
    _description = None

    @property
    def description(self):
        if self._description is None:
            classes = get_command_classes()
            cmds = []
            for name, cls in classes.items():
                cmds.append(_format_help(name, cls.short_description))
            self._description = (format_docstring(TEMPLATE % '\n'.join(cmds)) +
                "\n\nlymph help <command>     to display command specific help")
        return self._description

    def run(self):
        name = self.args['<command>']
        if name:
            print(get_command_class(name).get_help())
        else:
            print(self.description)




import unittest

import blessings
import gevent
import six

from lymph.cli.testing import CliIntegrationTestCase, CliTestMixin
from lymph.core.decorators import rpc
from lymph.core.interfaces import Interface


class Upper(Interface):

    @rpc()
    def upper(self, text=None):
        return text.upper()


class RequestCommandTests(CliIntegrationTestCase):
    def setUp(self):
        super(RequestCommandTests, self).setUp()
        self.upper_container, interface = self.create_container(Upper, 'upper')

    def test_request(self):
        result = self.cli(['request', 'upper.upper', '{"text":"foo"}'])
        self.assertEqual(result.returncode, 0)
        output = "u'FOO'" if six.PY2 else "'FOO'"
        self.assertEqual(result.stdout.rstrip(), output)

    def test_negative_request(self):
        result = self.cli(['request', 'no_exiting_container', '{}'])
        self.assertEqual(result.returncode, 1)

    def test_inspect(self):
        # Use --no-color to facilitate string comparison.
        result = self.cli(['inspect', '--no-color', 'upper'])
        self.assertEqual(result.returncode, 0)
        self.assertIn('upper.upper(text)', result.stdout)

    def test_negative_inspect(self):
        result = self.cli(['inspect', 'no_existing_container'])
        self.assertEqual(result.returncode, 1)


class ListCommandTests(CliTestMixin, unittest.TestCase):
    def test_list(self):
        self.assert_lines_equal(['list'], u"""
            {t.bold}tail       {t.normal}Stream the logs of one or more services.
            {t.bold}emit       {t.normal}Manually emits an event.
            {t.bold}request    {t.normal}Send a request message to some service and output the reply.
            {t.bold}inspect    {t.normal}Describes the RPC interface of a service
            {t.bold}discover   {t.normal}Show available services.
            {t.bold}help       {t.normal}Display help information about lymph.
            {t.bold}list       {t.normal}List available commands.
            {t.bold}subscribe  {t.normal}Prints events to stdout.
            {t.bold}instance   {t.normal}Run a single service instance (one process).
            {t.bold}node       {t.normal}Run a node service that manages a group of processes on the same machine.
            {t.bold}shell      {t.normal}Open an interactive Python shell locally or remotely.
        """.format(t=blessings.Terminal()), config=False)


class HelpCommandTests(CliTestMixin, unittest.TestCase):
    def test_help(self):
        self.assert_first_line_equals(['help'], 'Usage: lymph [options] <command> [<args>...]', config=False)

    def test_help_list(self):
        self.assert_first_line_equals(['help', 'list'], 'Usage: lymph list [options]', config=False)


class ServiceCommandTests(CliIntegrationTestCase):
    def test_instance(self):
        self.cli_config['interfaces'] = {
            'echo': {
                'class': 'lymph.tests.integration.test_cli:Upper',
            }
        }
        command_greenlet = gevent.spawn(self.cli, ['instance'])
        client = self.create_client()
        gevent.sleep(1)  # FIXME: how can we wait for the instance to register?
        response = client.request('echo', 'echo.upper', {'text': 'hi'}, timeout=1)
        self.assertEqual(response.body, 'HI')
        command_greenlet.kill()
        command_greenlet.join()



import sys
import os
import gevent


def watch_modules(callback):
    modules = {}
    while True:
        for name, module in list(sys.modules.items()):
            if module is None or not hasattr(module, '__file__'):
                continue
            module_source_path = os.path.abspath(module.__file__).rstrip('c')
            try:
                stat = os.stat(module_source_path)
            except OSError:
                continue
            mtime = stat.st_mtime
            if name in modules and modules[name] != mtime:
                callback()
            modules[name] = mtime
        gevent.sleep(1)


def set_source_change_callback(callback):
    gevent.spawn(watch_modules, callback)



import abc
import functools
import logging
import pkg_resources
import six
import textwrap

from lymph.exceptions import Timeout, LookupFailure

logger = logging.getLogger(__name__)


docstring_format_vars = {k: textwrap.dedent(v).strip() for k, v in six.iteritems({
    'COMMON_OPTIONS': """
Common Options:
  --config=<file>, -c <file>   Load configuration from the given path.
  --help, -h                   Print this help message and exit.
  --logfile=<file>             Redirect log output to the given file.
  --loglevel=<level>           Set the log level to one of DEBUG, INFO, WARNING,
                               ERROR. [default: WARNING]
  --version                    Show the lymph version and exit.
  --color                      Force colored output.
  --no-color                   Disable colored output.
  --vars=<file>                Load environment variables from the given path.
    """,
    'INSTANCE_OPTIONS': """
Instance Options:
  --isolated, -i               Don't register this service.
  --port=<port>, -p <port>     Use this port for the RPC endpoint.
  --ip=<address>               Use this IP for all sockets.
  --guess-external-ip, -g      Guess the public facing IP of this machine and
                               use it instead of the provided address.
  --reload                     Automatically stop the service when imported
                               python files in the current working directory
                               change. The process will be restarted by the
                               node. Do not use this in production.
    """,
})}


def format_docstring(doc):
    return textwrap.dedent(doc).format(**docstring_format_vars).strip()


@six.add_metaclass(abc.ABCMeta)
class Command(object):
    needs_config = True
    short_description = ''

    def __init__(self, args, config, terminal):
        self.args = args
        self.config = config
        self.terminal = terminal

    @classmethod
    def get_help(cls):
        return format_docstring(cls.__doc__)

    @abc.abstractmethod
    def run(self):
        raise NotImplementedError


_command_class_cache = None


def get_command_classes():
    global _command_class_cache
    if _command_class_cache is None:
        _command_class_cache, entry_points = {}, {}
        for entry_point in pkg_resources.iter_entry_points('lymph.cli'):
            name = entry_point.name
            if name in entry_points:
                logger.error('ignoring duplicate command definition for %s (already installed: %s)', entry_point, entry_points[name])
                continue
            entry_points[name] = entry_point
            cls = entry_point.load()
            cls.name = name
            _command_class_cache[name] = cls
    return _command_class_cache


def get_command_class(name):
    return get_command_classes()[name]


def handle_request_errors(func):
    @functools.wraps(func)
    def decorated(*args, **kwargs):
        try:
            func(*args, **kwargs)
        except LookupFailure as e:
            logger.error("The specified service name could not be found: %s: %s" % (type(e).__name__, e))
            return 1
        except Timeout:
            logger.error("The request timed out. Either the service is not available or busy.")
            return 1
    return decorated



import errno
import hashlib
import logging
import random
import time

import gevent
import zmq.green as zmq

from lymph.core.channels import RequestChannel, ReplyChannel
from lymph.core.components import Component
from lymph.core.connection import Connection
from lymph.core.messages import Message
from lymph.core.monitoring import metrics
from lymph.core.services import Service
from lymph.core import services
from lymph.core import trace
from lymph.exceptions import NotConnected


logger = logging.getLogger(__name__)


class ZmqRPCServer(Component):
    def __init__(self, ip='127.0.0.1', port=None, pool=None):
        super(ZmqRPCServer, self).__init__(pool=pool)
        self.ip = ip
        self.port = port

        self.zctx = zmq.Context.instance()
        self.endpoint = None
        self.bound = False
        self.request_counts = metrics.TaggedCounter('rpc')
        self.recv_loop_greenlet = None
        self.channels = {}
        self.connections = {}
        self.running = False
        self.request_handler = lambda channel: None

    @classmethod
    def from_config(cls, config, **kwargs):
        if 'pool' in config:
            pool = config.create_instance('pool', default_class='lymph.core.trace:Group')
        else:
            pool = None
        return cls(
            ip=config.get('ip', kwargs.get('ip') or '127.0.0.1'),
            port=config.get('port', kwargs.get('port')),
            pool=pool,
        )

    @property
    def identity(self):
        if self.endpoint:
            return hashlib.md5(self.endpoint.encode('utf-8')).hexdigest()

    def _bind(self, max_retries=2, retry_delay=0):
        assert not self.bound, 'already bound (endpoint=%s)' % self.endpoint
        self.send_sock = self.zctx.socket(zmq.ROUTER)
        self.recv_sock = self.zctx.socket(zmq.ROUTER)
        port = self.port
        retries = 0
        while True:
            if not self.port:
                port = random.randint(35536, 65536)
            try:
                self.endpoint = 'tcp://%s:%s' % (self.ip, port)
                endpoint = self.endpoint.encode('utf-8')
                self.recv_sock.setsockopt(zmq.IDENTITY, endpoint)
                self.send_sock.setsockopt(zmq.IDENTITY, endpoint)
                self.recv_sock.bind(self.endpoint)
            except zmq.ZMQError as e:
                if e.errno != errno.EADDRINUSE or retries >= max_retries:
                    raise
                logger.info('failed to bind to port %s (errno=%s), trying again.', port, e.errno)
                retries += 1
                if retry_delay:
                    gevent.sleep(retry_delay)
                continue
            else:
                self.port = port
                self.bound = True
                break

    def connect(self, endpoint):
        if endpoint not in self.connections:
            logger.debug("connecting to %s", endpoint)
            self.connections[endpoint] = Connection(self, endpoint)
            self.send_sock.connect(endpoint)
            gevent.sleep(0.02)
        return self.connections[endpoint]

    def disconnect(self, endpoint, socket=False):
        try:
            connection = self.connections[endpoint]
        except KeyError:
            return
        del self.connections[endpoint]
        connection.close()
        logger.debug("disconnecting from %s", endpoint)
        if socket:
            self.send_sock.disconnect(endpoint)

    def on_start(self):
        self._bind()
        self.running = True
        self.recv_loop_greenlet = self.spawn(self._recv_loop)

    def on_stop(self, **kwargs):
        self.running = False
        for connection in list(self.connections.values()):
            connection.close()
        if self.recv_loop_greenlet:
            self.recv_loop_greenlet.kill()
        self._close_sockets()

    def _close_sockets(self):
        self.recv_sock.close()
        self.send_sock.close()

    def _on_service_instance_unavailable(self, instance, action=None):
        self.disconnect(instance.endpoint)

    def _send_message(self, endpoint, msg):
        if not self.running:
            # FIXME: This should raise an Error instead of failing silently.
            logger.error('cannot send message (not started): %s', msg)
            return
        connection = self.connect(endpoint)
        self.send_sock.send(endpoint.encode('utf-8'), flags=zmq.SNDMORE)
        self.send_sock.send_multipart(msg.pack_frames())
        logger.debug('-> %s to %s', msg, endpoint)
        connection.on_send(msg)

    def prepare_headers(self, headers):
        headers = headers or {}
        headers.setdefault('trace_id', trace.get_id())
        return headers

    def _pick_endpoint(self, service):
        if not isinstance(service, Service):
            return service
        service.observe(services.REMOVED, self._on_service_instance_unavailable)
        choices = []
        for instance in service:
            try:
                connection = self.connections[instance.endpoint]
            except KeyError:
                choices.append(instance.endpoint)
                continue
            if connection.is_alive():
                choices.append(instance.endpoint)
        if not choices:
            raise NotConnected('Not connected to %s' % service.name)
        return random.choice(choices)

    def send_request(self, service, subject, body, headers=None):
        msg = Message(
            msg_type=Message.REQ,
            subject=subject,
            body=body,
            source=self.endpoint,
            headers=self.prepare_headers(headers),
        )
        channel = RequestChannel(msg, self)
        self.channels[msg.id] = channel
        try:
            endpoint = self._pick_endpoint(service)
        except NotConnected:
            logger.error('cannot send message (no instance): %s', msg)
        else:
            self._send_message(endpoint, msg)
        return channel

    def send_reply(self, msg, body, msg_type=Message.REP, headers=None):
        reply_msg = Message(
            msg_type=msg_type,
            subject=msg.id,
            body=body,
            source=self.endpoint,
            headers=self.prepare_headers(headers),
        )
        self._send_message(msg.source, reply_msg)
        return reply_msg

    def dispatch_request(self, msg):
        loglevel = self._get_loglevel(msg)
        logger.log(loglevel, '%s source=%s', msg.subject, msg.source)
        start = time.time()
        self.request_counts.incr(subject=msg.subject)
        channel = ReplyChannel(msg, self)
        try:
            self.request_handler(channel)
        finally:
            elapsed = time.time() - start
            logger.log(loglevel, 'subject=%s duration=%f (seconds)', msg.subject, elapsed)

    def _get_loglevel(self, msg):
        return logging.DEBUG if msg.subject == 'lymph.ping' else logging.INFO

    def recv_message(self, msg):
        trace.set_id(msg.headers.get('trace_id'))
        logger.debug('<- %s', msg)
        connection = self.connect(msg.source)
        connection.on_recv(msg)
        if msg.is_request():
            self.spawn(self.dispatch_request, msg)
        elif msg.is_reply():
            try:
                channel = self.channels[msg.subject]
            except KeyError:
                logger.debug('reply to unknown subject: %s (msg-id=%s)', msg.subject, msg.id)
                return
            channel.recv(msg)
        else:
            logger.warning('unknown message type: %s (msg-id=%s)', msg.type, msg.id)

    def _recv_loop(self):
        while True:
            frames = self.recv_sock.recv_multipart()
            try:
                msg = Message.unpack_frames(frames)
            except ValueError as e:
                msg_id = frames[1] if len(frames) >= 2 else None
                logger.warning('bad message format %s: %r (msg-id=%s)', e, (frames), msg_id)
                continue
            self.recv_message(msg)

    def ping(self, address):
        return self.send_request(address, 'lymph.ping', {'payload': ''})

    def _get_metrics(self):
        yield metrics.RawMetric('rpc.connection_count', len(self.connections))
        yield self.request_counts



import json

from werkzeug.exceptions import MethodNotAllowed


http_methods = ('get', 'post', 'head', 'options', 'put', 'patch', 'delete')


class RequestHandler(object):

    def __init__(self, interface, request):
        self.request = request
        self.interface = interface
        self._json = None

    @property
    def allowed_methods(self):
        return [method.upper() for method in http_methods if callable(getattr(self, method, None))]

    def json(self):
        if not "application/json" == self.request.mimetype:
            raise ValueError("The request Content-Type is not JSON")

        if self._json is None:
            self._json = json.loads(self.request.get_data(as_text=True))
        return self._json

    def dispatch(self, args):
        method = self.request.method.lower()
        if method not in http_methods:
            raise MethodNotAllowed(self.allowed_methods)
        try:
            func = getattr(self, method)
        except AttributeError:
            raise MethodNotAllowed(self.allowed_methods)
        return func(**args)



import json
import logging
import gevent
import os
import psutil
import six
from gevent import subprocess
from six.moves import range

from lymph.core.interfaces import Interface
from lymph.core.monitoring.metrics import RawMetric
from lymph.utils.sockets import create_socket


logger = logging.getLogger(__name__)


class Process(object):
    def __init__(self, cmd, env=None, service_type='n/a'):
        self.cmd = cmd
        self.env = env
        self.service_type = service_type
        self._process = None
        self._popen = None

    def is_running(self):
        return self._process and self._process.is_running()

    def start(self):
        self._popen = subprocess.Popen(
            self.cmd, env=self.env, close_fds=False)
        self._process = psutil.Process(self._popen.pid)

    def stop(self, **kwargs):
        signalnum = kwargs.get('signalnum')
        try:
            if signalnum:
                self._process.send_signal(signalnum)
            else:
                self._process.terminate()
            self._process.wait()
        except psutil.NoSuchProcess:
            pass

    def restart(self):
        print("restarting %s" % self)
        self.stop()
        self.start()

    def _get_metrics(self):
        if not self.is_running():
            return
        try:
            memory = self._process.memory_info()
            yield RawMetric('node.process.memory.rss', memory.rss, {'service_type': self.service_type})
            yield RawMetric('node.process.memory.vms', memory.vms, {'service_type': self.service_type})
            cpu = self._process.cpu_percent(interval=2.0)
            yield RawMetric('node.process.cpu', cpu, {'service_type': self.service_type})
        except psutil.NoSuchProcess:
            pass


class Node(Interface):
    register_with_coordinator = False

    def __init__(self, *args, **kwargs):
        super(Node, self).__init__(*args, **kwargs)
        self.sockets = {}
        self.processes = []
        self.running = False
        self._sockets = []
        self._services = []

    def apply_config(self, config):
        for name, c in six.iteritems(config.get('instances', {})):
            self._services.append((name, c.get('command'), c.get('numprocesses', 1)))

        socket_config = config.get_raw('sockets', ())
        if isinstance(socket_config, dict):
            socket_config = socket_config.values()
        for c in socket_config:
            self._sockets.append((c.get('host'), c.get('port')))

    def on_start(self):
        self.create_shared_sockets()
        self.running = True
        shared_fds = json.dumps({port: s.fileno() for port, s in six.iteritems(self.sockets)})
        for service_type, cmd, num in self._services:
            env = os.environ.copy()
            env.update({
                'LYMPH_NODE': self.container.endpoint,
                'LYMPH_MONITOR': self.container.monitor.endpoint,
                'LYMPH_NODE_IP': self.container.server.ip,
                'LYMPH_SHARED_SOCKET_FDS': shared_fds,
                'LYMPH_SERVICE_NAME': service_type,
            })
            for i in range(num):
                p = Process(cmd.split(' '), env=env, service_type=service_type)
                self.processes.append(p)
                logger.info('starting %s', cmd)
                p.start()
        self.container.spawn(self.watch_processes)

    def on_stop(self, **kwargs):
        logger.info("waiting for all service processes to die ...")
        self.running = False
        for p in self.processes:
            p.stop(**kwargs)
        super(Node, self).on_stop(**kwargs)

    def create_shared_sockets(self):
        for host, port in self._sockets:
            sock = create_socket(
                '%s:%s' % (host or self.container.server.ip, port), inheritable=True)
            self.sockets[port] = sock

    def restart_all(self):
        for process in self.processes:
            process.stop()

    def watch_processes(self):
        while True:
            for process in self.processes:
                try:
                    status = process._process.status()
                except psutil.NoSuchProcess:
                    if self.running:
                        process.start()
                    continue
                if status in (psutil.STATUS_ZOMBIE, psutil.STATUS_DEAD):
                    if self.running:
                        process.restart()
            gevent.sleep(1)

    def _get_metrics(self):
        for process in self.processes:
            for metric in process._get_metrics():
                yield metric



import unittest

import lymph
from lymph.core.interfaces import Interface
from lymph.core.messages import Message
from lymph.testing import RPCServiceTestCase
from lymph.exceptions import RemoteError, Nack


class Upper(Interface):
    service_type = 'upper'

    def __init__(self, *args, **kwargs):
        super(Upper, self).__init__(*args, **kwargs)
        self.eventlog = []

    @lymph.rpc()
    def upper(self, text=None):
        return text.upper()

    @lymph.rpc()
    def indirect_upper(self, text=None):
        # Method to test that it's possible to call upper method as
        # you do normally with any method.
        return self.upper(text)

    @lymph.rpc(raises=(ValueError,))
    def fail(self):
        raise ValueError('foobar')

    @lymph.raw_rpc()
    def just_ack(self, channel):
        channel.ack()

    @lymph.rpc()
    def auto_nack(self):
        raise ValueError('auto nack requested')

    @lymph.event('foo')
    def on_foo_event(self, event):
        self.eventlog.append((event.evt_type, event.body))


class BasicMockTest(RPCServiceTestCase):

    service_class = Upper
    service_name = 'upper'

    def test_upper(self):
        reply = self.client.upper(text='foo')
        self.assertEqual(reply, 'FOO')

    def test_indirect_upper(self):
        reply = self.client.indirect_upper(text='foo')
        self.assertEqual(reply, 'FOO')

    def test_ping(self):
        reply = self.request('lymph.ping', {'payload': 42})
        self.assertEqual(reply.body, 42)

    def test_status(self):
        reply = self.request('lymph.status', {})
        self.assertEqual(reply.body, {
            'endpoint': 'mock://300.0.0.1:1',
            'identity': '252946e723b6b07c1f5f0aa9442fb348',
        })

    def test_error(self):
        with self.assertRaisesRegexp(RemoteError, 'foobar'):
            self.client.fail()

    def test_exception_handling(self):
        with self.assertRaises(RemoteError.ValueError):
            self.client.fail()

    def test_ack(self):
        reply = self.request('upper.just_ack', {})
        self.assertIsNone(reply.body)
        self.assertEqual(reply.type, Message.ACK)

    def test_auto_nack(self):
        with self.assertRaises(Nack):
            self.client.auto_nack()

    def test_events(self):
        log = self.service.eventlog
        self.assertEqual(log, [])
        self.emit('foo', {'arg': 42})
        self.assertEqual(log, [('foo', {'arg': 42})])
        self.emit('foo', {'arg': 43})
        self.assertEqual(log, [('foo', {'arg': 42}), ('foo', {'arg': 43})])

    def test_inspect(self):
        proxy = self.get_proxy(namespace='lymph')
        methods = proxy.inspect()['methods']

        self.assertEqual(set(m['name'] for m in methods), set([
            'upper.fail', 'upper.upper', 'upper.auto_nack', 'upper.just_ack',
            'lymph.status', 'lymph.inspect', 'lymph.ping', 'upper.indirect_upper',
            'lymph.get_metrics',
        ]))



from __future__ import absolute_import

import logging
from logging.config import dictConfig

import six
import zmq.green as zmq

from lymph.utils.sockets import bind_zmq_socket


def get_loglevel(level_name):
    level = level_name.upper()
    if level not in ('DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'):
        raise ValueError("unknown loglevel: %s" % level)
    return getattr(logging, level)


def setup_logger(name):
    """Setup and return logger ``name`` with same settings as 'lymph' logger."""
    lymph_logger = logging.getLogger('lymph')
    logger = logging.getLogger(name)
    for hdlr in lymph_logger.handlers:
        logger.addHandler(hdlr)
    logger.setLevel(lymph_logger.level)
    # Since we are using DictConfig all logger are disabled by default first, so
    # we are enabling any logger here !
    logger.disabled = False
    return logger


class PubLogHandler(logging.Handler):
    def __init__(self, endpoint, socket=None):
        super(PubLogHandler, self).__init__()
        self.socket = socket
        if self.socket is None:
            ctx = zmq.Context.instance()
            self.socket = ctx.socket(zmq.PUB)
            endpoint, port = bind_zmq_socket(self.socket, endpoint)
        self.endpoint = endpoint

    def emit(self, record):
        topic = record.levelname
        self.socket.send_multipart([
            self._encode(topic),
            self._encode(self.endpoint),
            self._encode(self.format(record))])

    @staticmethod
    def _encode(potentially_text, encoding='utf-8'):
        if isinstance(potentially_text, six.text_type):
            return potentially_text.encode(encoding)
        return potentially_text


def setup_logging(config, loglevel, logfile):
    """Configure 'lymph' logger handlers and level.

    This function also set the container.log_endpoint in case it wasn't set.
    """
    logconf = dict(config.get('logging', {}))
    log_endpoint = config.get('container.log_endpoint')
    log_socket = None
    # Get log_endpoint in case it wasn't set in the config.
    if not log_endpoint:
        ctx = zmq.Context.instance()
        log_socket = ctx.socket(zmq.PUB)
        log_endpoint, port = bind_zmq_socket(log_socket, config.get('container.ip'))

        config.set('container.log_endpoint', log_endpoint)

    logconf.setdefault('version', 1)
    formatters = logconf.setdefault('formatters', {})
    formatters.setdefault('_trace', {
        '()': 'lymph.core.trace.TraceFormatter',
        'format': '%(asctime)s [%(levelname)s] %(name)s: %(message)s - trace_id="%(trace_id)s"',
    })
    handlers = logconf.setdefault('handlers', {})
    handlers.setdefault('_zmqpub', {
        'class': 'lymph.utils.logging.PubLogHandler',
        'formatter': '_trace',
        'endpoint': log_endpoint,
        'socket': log_socket,
    })
    console_logconf = {
        'class': 'logging.StreamHandler',
        'formatter': '_trace',
        'level': loglevel.upper(),
    }
    if logfile:
        console_logconf.update({
            'class': 'logging.FileHandler',
            'filename': logfile
        })
    handlers.setdefault('_console', console_logconf)
    loggers = logconf.setdefault('loggers', {})
    loggers.setdefault('lymph', {
        'handlers': ['_console', '_zmqpub'],
        'level': 'DEBUG',
    })
    dictConfig(logconf)



from raven import Client

from lymph.core.plugins import Plugin


class SentryPlugin(Plugin):
    def __init__(self, container, dsn=None, **kwargs):
        super(SentryPlugin, self).__init__()
        self.container = container
        self.client = Client(dsn)
        self.container.error_hook.install(self.on_error)

    def on_error(self, exc_info, **kwargs):
        self.client.captureException(exc_info, **kwargs)



import operator
import unittest

from lymph.core.monitoring import metrics


class RawMetricsTest(unittest.TestCase):

    def test_repr(self):
        raw = metrics.RawMetric('raw', 'foobar', {})

        self.assertEqual(repr(raw), "RawMetric(name='raw', value='foobar', tags={})")

    def test_iter(self):
        raw = metrics.RawMetric('raw', 'foobar', {})

        self.assertEqual(list(raw), [('raw', 'foobar', {})])


class CounterMetricsTest(unittest.TestCase):

    def test_get(self):
        counter = metrics.Counter('requests')

        self.assertEqual(list(counter), [('requests', 0, {})])

    def test_repr(self):
        counter = metrics.Counter('requests')

        self.assertEqual(str(counter), repr(counter))
        self.assertEqual(repr(counter), "Counter(name='requests', value=0, tags={})")

    def test_increment_by_one(self):
        counter = metrics.Counter('requests')
        counter += 1

        self.assertEqual(list(counter), [('requests', 1, {})])

    def test_increment_by_many(self):
        counter = metrics.Counter('requests')
        counter += 66

        self.assertEqual(list(counter), [('requests', 66, {})])


class TaggedCounterMetricsTest(unittest.TestCase):

    def test_incr_one_type(self):
        error_counter = metrics.TaggedCounter('exception')

        error_counter.incr(type='ValueError')
        error_counter.incr(4, type='ValueError')

        self.assertEqual(list(error_counter), [
            ('exception', 5, {'type': 'ValueError'})])

    def test_incr_different_types(self):
        error_counter = metrics.TaggedCounter('exception')

        error_counter.incr(type='ValueError')
        error_counter.incr(type='ValueError')

        error_counter.incr(type='Nack')

        self.assertEqual(
            sorted(error_counter, key=operator.itemgetter(1)), [
                ('exception', 1, {'type': 'Nack'}),
                ('exception', 2, {'type': 'ValueError'}),
            ])



from lymph.core.monitoring.global_metrics import RUsageMetrics, GeventMetrics, GarbageCollectionMetrics


class Aggregator(object):

    def __init__(self, *getters, **tags):
        self._metrics_getters = [
            RUsageMetrics(),
            GarbageCollectionMetrics(),
            GeventMetrics(),
        ] + list(getters)
        self._tags = tags

    def add_tags(self, **tags):
        self._tags.update(tags)

    def get_metrics(self):
        for getter in self._metrics_getters:
            for metrics in getter():
                for name, value, tags in metrics:
                    tags.update(self._tags)
                    yield name, value, tags



import logging
import time

import gevent
import msgpack
import zmq.green as zmq

from lymph.core.components import Component
from lymph.core.monitoring.global_metrics import RUsageMetrics, GeventMetrics, GarbageCollectionMetrics


logger = logging.getLogger(__name__)


DEFAULT_MONITOR_ENDPOINT = 'tcp://127.0.0.1:44044'


class MonitorPusher(Component):
    def __init__(self, container, aggregator, endpoint=None, interval=2):
        super(MonitorPusher, self).__init__()
        self.container = container
        self.interval = interval
        self.endpoint = endpoint or DEFAULT_MONITOR_ENDPOINT
        logger.info('connecting to monitor endpoint %s', self.endpoint)
        ctx = zmq.Context.instance()
        self.socket = ctx.socket(zmq.PUB)
        self.socket.connect(self.endpoint)

        self.aggregator = aggregator

    def on_start(self):
        self.loop_greenlet = self.container.spawn(self.loop)

    def on_stop(self, **kwargs):
        self.loop_greenlet.kill()

    def loop(self):
        last_stats = time.monotonic()
        while True:
            gevent.sleep(self.interval)
            dt = time.monotonic() - last_stats
            series = list(self.aggregator.get_metrics())
            stats = {
                'time': time.time(),
                'series': series,
            }
            last_stats += dt
            self.socket.send_multipart([b'stats', msgpack.dumps(stats)])



import unittest

from lymph.cli import base
from lymph.cli.testing import CommandFactory


class CommandFactoryTest(unittest.TestCase):

    def test_returns_a_command_instance(self):
        class ExampleCommand(base.Command):
            def run(self):
                pass

        _create_command = CommandFactory(ExampleCommand)
        cmd = _create_command()
        self.assertIsInstance(cmd, ExampleCommand)



from lymph.serializers import msgpack_serializer
from lymph.utils import make_id


class Message(object):
    ACK = b'ACK'
    REP = b'REP'
    REQ = b'REQ'
    NACK = b'NACK'
    ERROR = b'ERROR'

    def __init__(self, msg_type, subject, packed_body=None, headers=None, packed_headers=None, msg_id=None, source=None, lazy=False, **kwargs):
        self.id = msg_id if msg_id else make_id()
        self.type = msg_type
        self.subject = subject
        self.source = source

        if headers and packed_headers:
            raise TypeError("Message takes either 'headers' or 'packed_headers' not both")
        elif not headers and not packed_headers:
            headers = {}
        self._headers = headers
        self._packed_headers = packed_headers

        if 'body' in kwargs:
            if packed_body is not None:
                raise TypeError("Message takes either 'body' or 'packed_body' not both")
            self._body = kwargs['body']
        elif packed_body is None:
            raise TypeError("Message requires either 'body' or 'packed_body'")

        self._packed_body = packed_body
        if not lazy:
            self.body
            self.packed_body
            self.headers
            self.packed_headers

    def is_request(self):
        return self.type == self.REQ

    def is_reply(self):
        return self.type in (self.REP, self.ACK, self.NACK, self.ERROR)

    def is_idle_chatter(self):
        return not self.is_request() or self.subject == '_ping'

    @property
    def body(self):
        if not hasattr(self, '_body'):
            self._body = msgpack_serializer.loads(self._packed_body)
        return self._body

    @property
    def packed_body(self):
        if self._packed_body is None:
            self._packed_body = msgpack_serializer.dumps(self._body)
        return self._packed_body

    @property
    def headers(self):
        if self._headers is None:
            self._headers = msgpack_serializer.loads(self._packed_headers)
        return self._headers

    @property
    def packed_headers(self):
        if self._packed_headers is None:
            self._packed_headers = msgpack_serializer.dumps(self._headers)
        return self._packed_headers

    def pack_frames(self):
        return [
            self.id.encode('utf-8'),
            self.type,
            self.subject.encode('utf-8'),
            self.packed_headers,
            self.packed_body,
        ]

    @classmethod
    def unpack_frames(self, frames):
        try:
            source, msg_id, msg_type, subject, headers, body = frames
        except ValueError:
            raise ValueError('bad message frame count: got %s, expected 6' % len(frames))

        try:
            msg_id = msg_id.decode('utf-8')
            subject = subject.decode('utf-8')
            source = source.decode('utf-8')
        except UnicodeDecodeError:
            raise ValueError('message id, subject, and source must be utf-8 encoded.')

        return Message(
            msg_type=msg_type,
            subject=subject,
            msg_id=msg_id,
            source=source,
            packed_body=body,
            packed_headers=headers,
        )

    def __str__(self):
        return '{type=%s subject=%s id=%s..}' % (
            self.type,
            self.subject,
            self.id[:10],
        )

    def __repr__(self):
        return '<Message id=%s type=%s subject=%s body=%r>' % (
            self.id,
            self.type,
            self.subject,
            self.body,
        )



import datetime
import unittest
from lymph.utils.event_indexing import EventIndex
from lymph.core.events import Event
from mock import patch, Mock


class TestStore(unittest.TestCase):
    def setUp(self):
        def mockget(*args, **kwargs):
            return self.mocked
        def mockindex(*args, **kwargs):
            body = kwargs.get('body', {})
            self.mocked={'_source': body}
        self.es = Mock(get=mockget, index=mockindex)

    def test_stores_event(self):
        index = EventIndex(self.es, 'index_test_name')
        event = Event('test_event', {'number': 3,
                                     'string': 'hi',
                                     'float': 3.4,
                                     'dict': {'one': 1, 'two': 'dos'},
                                     'date': datetime.date(2014, 5, 2),
                                     'bool': True,
                                     'list': [1, 2, 3]})
        with patch('uuid.uuid4', Mock(hex='testuuid')):
            index.index(event)
        es_event = self.es.get(index='index_test_name', id='testuuid')
        self.assertEquals(es_event['_source']['s_string'], 'hi')
        self.assertEquals(es_event['_source']['i_number'], 3)
        self.assertEquals(es_event['_source']['f_float'], 3.4)
        self.assertEquals(es_event['_source']['o_dict'], {'s_two': 'dos',
                                                          'i_one': 1})
        self.assertEquals(es_event['_source']['d_date'], datetime.date(
            2014, 5, 2))
        self.assertEquals(es_event['_source']['b_bool'], True)
        self.assertEquals(es_event['_source']['l_list'], [1, 2, 3])
        self.assertEquals(es_event['_source']['type'], 'test_event')



import unittest

from lymph.core.monitoring.metrics import RawMetric
from lymph.core.monitoring.aggregator import Aggregator


def _get_metrics_one(self):
    yield RawMetric('dummy', 'one')


def _get_metrics_two(self):
    yield RawMetric('dummy', 'two')


class AggregatorTestCase(object):

    def test_aggregator_one_component(self):
        aggr = Aggregator(_get_metrics_one, name='test')

        self.assertEqual(list(aggr.get_metrics()), [('dummy', 'one', {'name': 'test'})])

    def test_aggregator_multiple_components(self):
        aggr = Aggregator(_get_metrics_one, _get_metrics_two, name='test')

        self.assertEqual(list(aggr.get_metrics()), [
            ('dummy', 'one', {'name': 'test'}),
            ('dummy', 'two', {'name': 'test'}),
        ])

    def test_aggregator_add_multiple_tags(self):
        aggr = Aggregator(_get_metrics_one, name='test')

        aggr.add_tags(origin='localhost', time='now')

        self.assertEqual(list(aggr.get_metrics()), [
            ('dummy', 'one', {'name': 'test', 'origin': 'localhost', 'time': 'now'})
        ])



# -*- coding: utf-8 -*-
import json

from lymph.client import Client
from lymph.cli.base import Command


class DiscoverCommand(Command):
    """
    Usage: lymph discover [--instances] [--ip=<address> | --guess-external-ip | -g] [--only-running] [options]

    Show available services

    Options:

      --instances                  Show service instances.
      --json                       Output json.
      --full                       Show all published instance meta data.
      --ip=<address>               Use this IP for all sockets.
      --guess-external-ip, -g      Guess the public facing IP of this machine and
                                   use it instead of the provided address.
      --only-running               Show only running services.

    {COMMON_OPTIONS}

    """

    short_description = 'Show available services.'

    def run(self):
        client = Client.from_config(self.config)
        services = client.container.discover()
        instances = {}
        for interface_name in sorted(services):
            service = client.container.lookup(interface_name)
            if not service and self.args.get('--only-running'):
                continue
            instances[interface_name] = service
        if self.args.get('--json'):
            print(json.dumps({
                name: [instance.serialize() for instance in service] for name, service in instances.items()
            }))
        else:
            self.print_human_readable_output(instances)

    def print_human_readable_output(self, instances):
        if instances:
            for interface_name, service in instances.items():
                print(u"%s [%s]" % (self.terminal.red(interface_name), len(service)))
                if self.args.get('--instances'):
                    service_instances = sorted(service, key=lambda d: d.identity)
                    for i, d in enumerate(service_instances):
                        prefix = u'└─' if i == len(service_instances) - 1 else u'├─'
                        print(u'%s [%s] %s' % (prefix, d.identity[:10], d.endpoint))
                        if self.args.get('--full'):
                            for k, v in sorted(d.serialize().items()):
                                if k == 'endpoint':
                                    continue
                                print(u'   %s: %r' % (k, v))
        else:
            print(u"No registered services found")




class Observable(object):
    def __init__(self):
        self.observers = {}

    def notify_observers(self, action, *args, **kwargs):
        kwargs.setdefault('action', action)
        for callback in self.observers.get(action, ()):
            callback(*args, **kwargs)

    def observe(self, actions, callback):
        if not isinstance(actions, (tuple, list)):
            actions = (actions,)
        for action in actions:
            self.observers.setdefault(action, set()).add(callback)



import six


class RpcError(Exception):
    def __repr__(self):
        return '<%s: %s>' % (self.__class__.__name__, self)


class RpcRequestError(RpcError):
    def __init__(self, request, *args, **kwargs):
        self.request = request
        super(RpcError, self).__init__(*args, **kwargs)


class Timeout(RpcRequestError):
    pass


class Nack(RpcRequestError):
    pass


class LookupFailure(RpcError):
    pass


class RegistrationFailure(RpcError):
    pass


class EventHandlerTimeout(Exception):
    pass


class _RemoteException(type):

    # Hold dynamically generated exception classes.
    __exclasses = {}

    def __getattr__(cls, errtype):
        return cls.__exclasses.setdefault(errtype, type(errtype, (cls,), {}))


@six.add_metaclass(_RemoteException)
class RemoteError(RpcRequestError):
    @classmethod
    def from_reply(cls, request, reply):
        errtype = reply.body.get('type', cls.__name__)
        subcls = getattr(cls, errtype)
        return subcls(request, reply.body.get('message', ''))


class SocketNotCreated(Exception):
    pass


class NotConnected(Exception):
    pass


class ResourceExhausted(Exception):
    pass


class ConfigurationError(Exception):
    pass



# -*- coding: utf-8 -*-
from __future__ import print_function
import json
import logging
import math
import pprint
import sys
import time

from gevent.pool import Pool

from lymph.client import Client
from lymph.exceptions import Timeout
from lymph.cli.base import Command, handle_request_errors


logger = logging.getLogger(__name__)


class RequestCommand(Command):
    """
    Usage: lymph request [options] <subject> <params>

    Description:
        Sends a single RPC request to <address>. Parameters have to be JSON encoded.

    Options:
      --ip=<address>               Use this IP for all sockets.
      --guess-external-ip, -g      Guess the public facing IP of this machine and
                                   use it instead of the provided address.
      --timeout=<seconds>          RPC timeout. [default: 2.0]
      --address=<addr>             Send the request to the given instance.
      -N <number>                  Send a total of <N> requests [default: 1].
      -C <concurrency>             Send requests from <concurrency> concurrent greenlets [default: 1].

    {COMMON_OPTIONS}
    """

    short_description = 'Send a request message to some service and output the reply.'

    def _run_one_request(self, request):
        pprint.pprint(request().body)

    def _run_many_requests(self, request, n, c):
        # one warm up request for lookup and connection creation
        request()

        timings = []
        timeouts = []

        def timed_request(i):
            start = time.time()
            try:
                request()
            except Timeout:
                timeouts.append(i)
            else:
                timings.append(1000 * (time.time() - start))
            request_count = len(timings) + len(timeouts)
            if request_count % (n / 80) == 0:
                sys.stdout.write('.')
                sys.stdout.flush()

        pool = Pool(size=c)
        print("sending %i requests, concurrency %i" % (n, c))
        start = time.time()
        pool.map(timed_request, range(n))
        total_time = (time.time() - start)

        timings.sort()
        n_success = len(timings)
        n_timeout = len(timeouts)
        avg = sum(timings) / n_success
        stddev = math.sqrt(sum((t - avg) ** 2 for t in timings)) / n_success

        print()
        print('Requests per second:   %8.2f Hz  (#req=%s)' % (n_success / total_time, n_success))
        print('Mean time per request: %8.2f ms  (stddev=%.2f)' % (avg, stddev))
        print('Timeout rate:          %8.2f %%   (#req=%s)' % (100 * n_timeout / float(n), n_timeout))
        print('Total time:            %8.2f s' % total_time)
        print()

        print('Percentiles:')
        print('  0.0 %%   %8.2f ms (min)' % timings[0])
        for p in (50, 90, 95, 97, 98, 99, 99.5, 99.9):
            print('%5.1f %%   %8.2f ms' % (p, timings[int(math.floor(0.01 * p * n_success))]))
        print('100.0 %%   %8.2f ms (max)' % timings[-1])

    @handle_request_errors
    def run(self):
        body = json.loads(self.args.get('<params>', '{}'))
        try:
            timeout = float(self.args.get('--timeout'))
        except ValueError:
            print("--timeout requires a number (e.g. --timeout=0.42)")
            return 1
        subject = self.args['<subject>']
        address = self.args.get('--address')
        if not address:
            address = subject.split('.', 1)[0]

        client = Client.from_config(self.config)

        def request():
            return client.request(address, subject, body, timeout=timeout)

        N, C = int(self.args['-N']), int(self.args['-C'])

        if N == 1:
            return self._run_one_request(request)
        else:
            return self._run_many_requests(request, N, C)




import unittest

from lymph.core.events import EventDispatcher


class EventDispatcherTest(unittest.TestCase):
    def setUp(self):
        self.dispatcher = EventDispatcher()
        self.handler_log = []
        self.handlers = {}

    def make_handler(self, name):
        if name in self.handlers:
            return self.handlers[name]

        def handler(*args):
            self.handler_log.append((name, args))

        handler.__name__ = name
        self.handlers[name] = handler
        return handler

    def assert_dispatched_patterns_equal(self, event_type, patterns):
        self.assertEqual(
            set(pattern for pattern, handler in self.dispatcher.dispatch(event_type)),
            set(patterns),
        )

    def assert_dispatched_handlers_equal(self, event_type, handlers):
        self.assertEqual(
            set(handler.__name__ for pattern, handler in self.dispatcher.dispatch(event_type)),
            set(handlers),
        )

    def test_basic_dispatch(self):
        self.dispatcher.register('foo', self.make_handler('foo'))
        self.dispatcher.register('bar', self.make_handler('bar'))
        self.dispatcher.register('foo.bar', self.make_handler('foo2'))

        self.assert_dispatched_patterns_equal('foo', {'foo'})
        self.assert_dispatched_patterns_equal('bar', {'bar'})
        self.assert_dispatched_patterns_equal('fooo', [])
        self.assert_dispatched_patterns_equal('foofoo', [])
        self.assert_dispatched_patterns_equal('', [])

    def test_wildcard_dispatch(self):
        self.dispatcher.register('foo', self.make_handler('foo'))
        self.dispatcher.register('#', self.make_handler('hash'))
        self.dispatcher.register('*', self.make_handler('star'))
        self.dispatcher.register('foo.*', self.make_handler('foo_star'))
        self.dispatcher.register('foo.#', self.make_handler('foo_hash'))

        self.assert_dispatched_patterns_equal('foo', {'foo', '*', '#'})
        self.assert_dispatched_patterns_equal('foo.bar', {'#', 'foo.*', 'foo.#'})
        self.assert_dispatched_patterns_equal('foo.bar.baz', {'#', 'foo.#'})
        self.assert_dispatched_patterns_equal('', {'#'})

    def test_multi_pattern_registration(self):
        self.dispatcher.register('foo', self.make_handler('foo'))
        self.dispatcher.register('#', self.make_handler('foo'))

        self.assert_dispatched_patterns_equal('foo', {'foo', '#'})
        self.assert_dispatched_handlers_equal('foo', {'foo'})

    def test_update(self):
        self.dispatcher.register('foo', self.make_handler('base_foo'))
        self.dispatcher.register('#', self.make_handler('hash'))

        ed = EventDispatcher()
        ed.register('foo', self.make_handler('foo'))
        ed.register('bar', self.make_handler('bar'))
        self.dispatcher.update(ed)

        self.assert_dispatched_handlers_equal('foo', {'foo', 'base_foo', 'hash'})
        self.assert_dispatched_handlers_equal('bar', {'hash', 'bar'})
